"""
Code for a class to calculate receiver operating characteristic (ROC) curves.
"""
from pathlib import Path
from typing import Union

import numpy as np
import matplotlib.pyplot as plt


class RocCalculator:
    """
    A class to calculate receiver operating characteristic (ROC) curves.
    """

    def __init__(self):
        self.true_positive_counts: int = 0
        self.false_positive_counts: int = 0
        self.true_negative_counts: int = 0
        self.false_negative_counts: int = 0
        self.thresholds: np.ndarray = np.linspace(0, 1, 1000)

    @property
    def true_positive_rates(self) -> np.ndarray:
        """
        Calculates the true positive rates for the accumulated confusion matrix counts for each threshold.

        :return: The true positive rates for each threshold.
        """
        return self.true_positive_counts / (self.true_positive_counts + self.false_negative_counts)

    @property
    def false_positive_rates(self) -> np.ndarray:
        """
        Calculates the false positive rates for the accumulated confusion matrix counts for each threshold.

        :return: The false positive rates for each threshold.
        """
        return self.false_positive_counts / (self.true_negative_counts + self.false_positive_counts)

    @staticmethod
    def threshold_predictions(probability_predictions: np.ndarray, thresholds: np.ndarray):
        """
        From a 1D array of probability predictions, calculates a 2D array of binary predictions with each row
        corresponding to the predictions given one of the passed probability thresholds.

        :param probability_predictions: The array of predicted probabilities for the binary labels.
        :param thresholds: The thresholds to generate binary labels on from the probabilities.
        :return: The array containing the binary labels for each threshold.
        """
        return probability_predictions > thresholds[:, np.newaxis]

    @staticmethod
    def calculate_confusion_matrix_counts(label: np.ndarray, predictions: np.ndarray):
        """
        Calculates the confusion matrix counts for a 1D set of true binary labels a 2D array of predictions, where
        each row corresponds to a prediction to compare.

        :param label: A 1D binary array label.
        :param predictions: A 2D array of predictions, each row of which is to be compared to the label.
        :return: The confusion matrix values for each row of the predictions.
        """
        true_positives = np.count_nonzero(label & predictions, axis=1)
        false_positives = np.count_nonzero(~label & predictions, axis=1)
        true_negatives = np.count_nonzero(~label & ~predictions, axis=1)
        false_negatives = np.count_nonzero(label & ~predictions, axis=1)
        return true_positives, false_positives, true_negatives, false_negatives

    def accumulate_confusion_matrix_counts(self, label: np.ndarray, prediction: np.ndarray):
        """
        Calculates the confusion matrix counts for a given label and probability prediction pair, and adds those counts
        to the totals.

        :param label: The 1D array binary label.
        :param prediction: The 1D probability array prediction.
        """
        thresholded_predictions = self.threshold_predictions(prediction, self.thresholds)
        counts = self.calculate_confusion_matrix_counts(label, thresholded_predictions)
        true_positives, false_positives, true_negatives, false_negatives = counts
        self.true_positive_counts += true_positives
        self.false_positive_counts += false_positives
        self.true_negative_counts += true_negatives
        self.false_negative_counts += false_negatives

    def generate_roc_plot(self, output_path: Union[str, Path] = 'roc_curve.svg', title: str = None):
        """
        Generates a ROC curve plot from the confusion matrix totals which have been accumulated.

        :param output_path: The path to save the plot image file to.
        :param title: The title to add to the plot.
        """
        with plt.style.context('seaborn-darkgrid'):
            figure, axes = plt.subplots()
            axes.plot(self.false_positive_rates, self.true_positive_rates)
            axes.set_xlabel('False positive rate')
            axes.set_ylabel('True positive rate')
            if title is not None:
                axes.set_title(title)
            figure.patch.set_alpha(0)  # Transparent figure background while keeping grid background.
            # Need to explicitly state face color or the save will override it.
            plt.savefig(output_path, facecolor=figure.get_facecolor())



"""Code grouping the objects related to one entity for the viewer."""
import asyncio
from pathlib import Path
from typing import Any, Union

from ramjet.analysis.transit_vetter import TransitVetter
from ramjet.data_interface.tess_toi_data_interface import TessToiDataInterface
from ramjet.photometric_database.tess_ffi_light_curve import TessFfiLightCurve
from ramjet.photometric_database.tess_light_curve import TessLightCurve
from ramjet.photometric_database.tess_target import TessTarget


class ViewEntity:
    """A class grouping the objects related to one entity for the viewer."""
    tess_toi_data_interface = TessToiDataInterface()
    vetter = TransitVetter()

    def __init__(self):
        self.index: Union[int, None] = None
        self.confidence: Union[float, None] = None
        self.light_curve: Union[TessLightCurve, None] = None
        self.target: Union[TessTarget, None] = None
        self.has_exofop_dispositions: Union[bool, None] = None
        self.has_nearby_toi: Union[bool, None] = None

    @classmethod
    async def from_identifier_data_frame_row(cls, identifier_data_frame_row):
        """
        Creates the view entity based on a data frame row from an infer output.

        :param identifier_data_frame_row: The row of the data frame that should be used to prepare the view entity.
        :return: The view entity.
        """
        view_entity = cls()
        loop = asyncio.get_running_loop()
        light_curve_path_string = identifier_data_frame_row['light_curve_path']
        light_curve_path = Path(light_curve_path_string)
        load_light_curve_task = loop.run_in_executor(None, view_entity.load_light_curve_from_identifier,
                                                     light_curve_path)
        tic_id, sector = TessFfiLightCurve.get_tic_id_and_sector_from_file_path(light_curve_path)
        load_has_dispositions_task = loop.run_in_executor(
            None, view_entity.tess_toi_data_interface.has_any_exofop_dispositions_for_tic_id, tic_id
        )
        tic_id_only_target = TessTarget()  # Create a stand-in target to use for other parallel preloading.
        tic_id_only_target.tic_id = tic_id
        load_nearby_toi_task = loop.run_in_executor(
            None, view_entity.vetter.has_nearby_toi_targets, tic_id_only_target
        )
        load_target_task = loop.run_in_executor(None, TessTarget.from_tic_id, tic_id)
        light_curve, target, has_dispositions, has_nearby_toi = await asyncio.gather(
            load_light_curve_task, load_target_task, load_has_dispositions_task, load_nearby_toi_task)
        view_entity.has_nearby_toi = has_nearby_toi
        view_entity.has_exofop_dispositions = has_dispositions
        view_entity.index = identifier_data_frame_row['index']
        view_entity.confidence = identifier_data_frame_row['confidence']
        view_entity.light_curve = light_curve
        view_entity.target = target
        return view_entity

    @staticmethod
    def load_light_curve_from_identifier(identifier: Any) -> TessLightCurve:
        """
        Loads a light curve from a generic identifier.

        :param identifier: The identifier of the light curve.
        :return: The light curve.
        """
        light_curve = TessFfiLightCurve.from_path(identifier)
        light_curve.convert_to_relative_scale()
        return light_curve



"""
Code to load view entities in the background so they show up quickly when displayed.
"""
import pandas as pd
import asyncio
from asyncio import Task
from collections import deque
import warnings
from pathlib import Path
from typing import Union, Deque

from ramjet.analysis.viewer.view_entity import ViewEntity
from ramjet.data_interface.tess_data_interface import NoDataProductsFoundException


class Preloader:
    """
    A class to load view entities in the background so they show up quickly when displayed.
    """
    minimum_preloaded = 25
    maximum_preloaded = 50

    def __init__(self):
        self.current_view_entity: Union[None, ViewEntity] = None
        self.next_view_entity_deque: Deque[ViewEntity] = deque(maxlen=self.maximum_preloaded)
        self.previous_view_entity_deque: Deque[ViewEntity] = deque(maxlen=self.maximum_preloaded)
        self.identifier_data_frame: Union[pd.DataFrame, None] = None
        self.running_loading_task: Union[Task, None] = None

    async def load_view_entity_at_index_as_current(self, index: int):
        """
        Loads the view entity at the passed index as the current view entity.

        :param index: The index in the path list to load.
        """
        await self.cancel_loading_task()
        self.current_view_entity = await ViewEntity.from_identifier_data_frame_row(
            self.identifier_data_frame.iloc[index])
        await self.reset_deques()

    async def load_surrounding_view_entities(self):
        """
        Loads the next and previous view entities relative to the current view entity.
        """
        await self.load_next_view_entities()
        await self.load_previous_view_entities()

    async def load_next_view_entities(self):
        """
        Preload the next view entities.
        """
        if len(self.next_view_entity_deque) > 0:
            last_index = self.next_view_entity_deque[-1].index
        else:
            last_index = self.current_view_entity.index
        while (len(self.next_view_entity_deque) < self.minimum_preloaded and
               last_index != self.identifier_data_frame.shape[0] - 1):
            last_index += 1
            try:
                last_view_entity = await ViewEntity.from_identifier_data_frame_row(
                    self.identifier_data_frame.iloc[last_index])
            except NoDataProductsFoundException:
                warnings.warn(f'No light curve found for identifier {self.identifier_data_frame.iloc[last_index]}.')
                continue
            self.next_view_entity_deque.append(last_view_entity)

    async def load_previous_view_entities(self):
        """
        Preload the previous view entities.
        """
        if len(self.previous_view_entity_deque) > 0:
            first_index = self.previous_view_entity_deque[0].index
        else:
            first_index = self.current_view_entity.index
        while (len(self.previous_view_entity_deque) < self.minimum_preloaded and
               first_index != 0):
            first_index -= 1
            try:
                first_view_entity = await ViewEntity.from_identifier_data_frame_row(
                    self.identifier_data_frame.iloc[first_index])
            except NoDataProductsFoundException:
                warnings.warn(f'No light curve found for identifier {self.identifier_data_frame.iloc[first_index]}.')
                continue
            self.previous_view_entity_deque.appendleft(first_view_entity)

    async def increment(self) -> ViewEntity:
        """
        Increments to the next view entity, and calls loading as necessary.

        :return: The new current view entity.
        """
        self.previous_view_entity_deque.append(self.current_view_entity)
        if len(self.next_view_entity_deque) == 0 and (
                self.running_loading_task is not None and not self.running_loading_task.done()):
            await self.running_loading_task
        self.current_view_entity = self.next_view_entity_deque.popleft()
        await self.refresh_surrounding_light_curve_loading()
        return self.current_view_entity

    async def decrement(self) -> ViewEntity:
        """
        Decrements to the previous view entity, and calls loading as necessary.

        :return: The new current view entity.
        """
        self.next_view_entity_deque.appendleft(self.current_view_entity)
        while len(self.previous_view_entity_deque) == 0 and (
                self.running_loading_task is not None and not self.running_loading_task.done()):
            await self.running_loading_task
        self.current_view_entity = self.previous_view_entity_deque.pop()
        await self.refresh_surrounding_light_curve_loading()
        return self.current_view_entity

    async def refresh_surrounding_light_curve_loading(self):
        """
        Cancels the existing loading task and starts a new one.
        """
        await self.cancel_loading_task()
        self.running_loading_task = asyncio.create_task(self.load_surrounding_view_entities())

    async def cancel_loading_task(self):
        """
        Cancels an existing loading task if it exists.
        """
        if self.running_loading_task is not None:
            self.running_loading_task.cancel()
            try:
                await self.running_loading_task
            except asyncio.CancelledError:
                pass

    async def reset_deques(self):
        """
        Cancels any loading tasks, clears the deques, and starts the loading task.
        """
        await self.cancel_loading_task()
        self.previous_view_entity_deque = deque(maxlen=self.maximum_preloaded)
        self.next_view_entity_deque = deque(maxlen=self.maximum_preloaded)
        self.running_loading_task = asyncio.create_task(self.load_surrounding_view_entities())

    @classmethod
    async def from_csv_path(cls, csv_path: Path, starting_index: int = 0):
        """
        Create a preloader from a CSV of light curve identifiers.

        :param csv_path: A path to a CSV containing light curve identifier information.
        :param starting_index: The starting index to preload around.
        :return: The preloader.
        """
        preloader = cls()
        preloader.identifier_data_frame = pd.read_csv(csv_path)
        await preloader.load_view_entity_at_index_as_current(starting_index)
        return preloader



from dataclasses import dataclass
from random import Random
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterable, Callable, Self, Iterator, List, Tuple

import numpy as np
import numpy.typing as npt

from qusi.light_curve import LightCurve
from qusi.light_curve_observation import LightCurveObservation


class LightCurveCollectionBase(ABC):
    @abstractmethod
    def light_curve_iter(self) -> Iterator[LightCurve]:
        pass

    @abstractmethod
    def load_times_and_fluxes_from_path(self, path) -> Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]:
        pass


class LightCurveObservationCollectionBase(LightCurveCollectionBase):
    @abstractmethod
    def observation_iter(self) -> Iterator[LightCurveObservation]:
        pass

class LightCurveObservationIndexableBase(ABC):
    @abstractmethod
    def __getitem__(self, indexes: int | Tuple[int]) -> LightCurveObservation | Tuple[LightCurveObservation]:
        pass

class PathIterableBase(ABC):
    @abstractmethod
    def get_shuffled_paths(self) -> List[Path]:
        pass

    @abstractmethod
    def get_paths(self) -> List[Path]:
        pass


class PathIndexableBase(ABC):
    @abstractmethod
    def __getitem__(self, indexes: int | Tuple[int]) -> Path | Tuple[Path]:
        pass


class PathGetterBase(PathIterableBase, PathIndexableBase):
    pass


@dataclass
class PathGetter(PathGetterBase):
    """
    A class to get paths from a path generation function.

    :ivar get_paths_function: The function which returns the path iterable.
    :ivar random_number_generator: A random number generator.
    """
    get_paths_function: Callable[[], Iterable[Path]]
    random_number_generator: Random

    @classmethod
    def new(cls, get_paths_function: Callable[[], Iterable[Path]]) -> Self:
        random_number_generator = Random(0)
        instance = cls(get_paths_function=get_paths_function, random_number_generator=random_number_generator)
        return instance

    def get_shuffled_paths(self) -> Iterable[Path]:
        """
        Gets the shuffled paths iterable.

        :return: The shuffled paths iterable.
        """
        light_curve_paths = self.get_paths()
        self.random_number_generator.shuffle(light_curve_paths)
        return light_curve_paths

    def get_paths(self):
        """
        Gets the paths iterable.

        :return: The paths iterable.
        """
        light_curve_paths = list(self.get_paths_function())
        return light_curve_paths

    def __getitem__(self, index: int | Tuple[int]) -> Path | Tuple[Path]:
        light_curve_paths = self.get_paths()
        indexed_light_curve_paths = np.array(light_curve_paths)[index]
        if isinstance(indexed_light_curve_paths, Path):
            return indexed_light_curve_paths
        else:
            return indexed_light_curve_paths.tolist()


@dataclass
class LightCurveCollection(LightCurveCollectionBase, LightCurveObservationIndexableBase):
    """
    :ivar path_getter: The PathIterableBase object for the collection.
    :ivar load_times_and_fluxes_from_path_function: The function to load the times and fluxes from the light curve.
    """
    path_getter: PathGetterBase
    load_times_and_fluxes_from_path_function: Callable[
        [Path], Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]]

    @classmethod
    def new(cls,
            get_paths_function: Callable[[], Iterable[Path]],
            load_times_and_fluxes_from_path_function: Callable[
                [Path], Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]],
            ) -> Self:
        """
        Creates a new light curve collection.

        :param get_paths_function: The function to load the list of paths.
        :param load_times_and_fluxes_from_path_function: The function to load the times and fluxes from the light curve.
        :return: The light curve collection.
        """
        path_getter = PathGetter.new(get_paths_function=get_paths_function)
        return cls(path_getter=path_getter,
                   load_times_and_fluxes_from_path_function=load_times_and_fluxes_from_path_function)

    def light_curve_iter(self) -> Iterator[LightCurve]:
        """
        Get the iterable that will iterate through the light curves of the collection.

        :return: The iterable of the light curves.
        """
        light_curve_paths = self.path_getter.get_shuffled_paths()
        for light_curve_path in light_curve_paths:
            times, fluxes = self.load_times_and_fluxes_from_path_function(light_curve_path)
            light_curve = LightCurve.new(times, fluxes)
            yield light_curve

    def load_times_and_fluxes_from_path(self, path) -> Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]:
        return self.load_times_and_fluxes_from_path_function(path)

    def __getitem__(self, index: int) -> LightCurve:
        light_curve_path = self.path_getter[index]
        times, fluxes = self.load_times_and_fluxes_from_path(light_curve_path)
        light_curve = LightCurve.new(times, fluxes)
        return light_curve


@dataclass
class LabeledLightCurveCollection(LightCurveObservationCollectionBase, LightCurveObservationIndexableBase):
    """
    :ivar path_getter: The PathGetterBase object for the collection.
    :ivar light_curve_collection: The LightCurveCollectionBase object for the collection.
    :ivar load_label_from_path_function: The function to load the label for the light curve.
    """
    path_getter: PathGetterBase
    light_curve_collection: LightCurveCollectionBase
    load_label_from_path_function: Callable[[Path], int]

    @classmethod
    def new(cls,
            get_paths_function: Callable[[], Iterable[Path]],
            load_times_and_fluxes_from_path_function: Callable[
                [Path], Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]],
            load_label_from_path_function: Callable[[Path], int]
            ) -> Self:
        """
        Creates a new light curve collection.

        :param get_paths_function: The function to load the list of paths.
        :param load_times_and_fluxes_from_path_function: The function to load the times and fluxes from the light curve.
        :param load_label_from_path_function: The function to load the label for the light curve.
        :return: The light curve collection.
        """
        path_iterable = PathGetter.new(get_paths_function=get_paths_function)
        light_curve_collection = LightCurveCollection(
            path_getter=path_iterable,
            load_times_and_fluxes_from_path_function=load_times_and_fluxes_from_path_function)
        return cls(path_getter=path_iterable,
                   light_curve_collection=light_curve_collection,
                   load_label_from_path_function=load_label_from_path_function)

    @classmethod
    def new_with_label(cls,
                       get_paths_function: Callable[[], Iterable[Path]],
                       load_times_and_fluxes_from_path_function: Callable[
                           [Path], Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]],
                       label: int
                       ) -> Self:
        """
        Creates a new light curve collection with a specific label for all light curves..

        :param get_paths_function: The function to load the list of paths.
        :param load_times_and_fluxes_from_path_function: The function to load the times and fluxes from the light curve.
        :param label: The label to be applied to all light curves in the collection.
        :return: The light curve collection.
        """
        load_label_from_path_function = create_constant_label_for_path_function(label)
        path_iterable = PathGetter.new(get_paths_function=get_paths_function)
        light_curve_collection = LightCurveCollection(
            path_getter=path_iterable,
            load_times_and_fluxes_from_path_function=load_times_and_fluxes_from_path_function)
        return cls(path_getter=path_iterable,
                   light_curve_collection=light_curve_collection,
                   load_label_from_path_function=load_label_from_path_function)

    def light_curve_iter(self) -> Iterator[LightCurve]:
        return self.light_curve_collection.light_curve_iter()

    def load_times_and_fluxes_from_path(self, path) -> Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]:
        return self.light_curve_collection.load_times_and_fluxes_from_path(path=path)

    def observation_iter(self) -> Iterator[LightCurveObservation]:
        """
        Get the iterable that will iterate through the light curves of the collection.

        :return: The iterable of the light curves.
        """
        light_curve_paths = self.path_getter.get_shuffled_paths()
        for light_curve_path in light_curve_paths:
            times, fluxes = self.light_curve_collection.load_times_and_fluxes_from_path(light_curve_path)
            label = self.load_label_from_path_function(light_curve_path)
            light_curve = LightCurve.new(times, fluxes)
            light_curve_observation = LightCurveObservation.new(light_curve, label)
            yield light_curve_observation

    def __getitem__(self, index: int) -> LightCurveObservation:
        light_curve_path = self.path_getter[index]
        times, fluxes = self.light_curve_collection.load_times_and_fluxes_from_path(light_curve_path)
        label = self.load_label_from_path_function(light_curve_path)
        light_curve = LightCurve.new(times, fluxes)
        light_curve_observation = LightCurveObservation.new(light_curve, label)
        return light_curve_observation


def create_constant_label_for_path_function(label: int) -> Callable[[Path], int]:
    """
    Creates a closure function that accepts a path, but always returns a given label regardless of the path.

    :param label: The label that the closure function should return.
    :return: The closure function.
    """

    def constant_label_for_path(_path: Path) -> int:
        """
        The function which will return the outer label regardless of path.

        :param _path: The unused path.
        :return: The label
        """
        return label

    return constant_label_for_path



"""
Code for visualizing light curves.
"""

from pathlib import Path
from typing import Union

import numpy
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from bokeh.models import LinearColorMapper
from bokeh.palettes import Turbo256
from bokeh.plotting import figure as Figure
from matplotlib.colors import LinearSegmentedColormap


def plot_light_curve(times: np.ndarray, fluxes: np.ndarray, labels: np.ndarray = None, predictions: np.ndarray = None,
                     title: str = None, x_label: str = 'Days', y_label: str = 'Flux',
                     x_limits: (float, float) = (None, None), y_limits: (float, float) = (None, None),
                     save_path: Union[Path, str] = None, exclude_flux_outliers: bool = False,
                     base_data_point_size: float = 3):
    """
    Plots a light curve with a consistent styling. If true labels and/or predictions are included, these will
    additionally be plotted.

    :param times: The times of the measurements.
    :param fluxes: The fluxes of the measurements.
    :param labels: The binary labels for each time step.
    :param predictions: The probability prediction for each time step.
    :param title: The title of the plot.
    :param x_label: The label for the x axis.
    :param y_label: The label for the y axis.
    :param x_limits: Optional axis limiting for the x axis.
    :param y_limits: Optional axis limiting for the y axis.
    :param save_path: The path to save the plot to. If `None`, the plot will be shown instead.
    :param exclude_flux_outliers: Whether or not to exclude flux outlier data points when plotting.
    :param base_data_point_size: The size of the data points to use when plotting (and related sizes).
    """
    with plt.style.context('seaborn-whitegrid'):
        figure, axes = plt.subplots(figsize=(16, 10))
        axes.set_xlabel(x_label)
        axes.set_ylabel(y_label)
        color_map = plt.get_cmap('tab10')
        data_point_color = color_map(0)
        positive_data_point_color = color_map(2)
        prediction_color = color_map(3)
        if exclude_flux_outliers:
            outlier_indices = is_outlier(fluxes)
            fluxes = fluxes[~outlier_indices]
            times = times[~outlier_indices]
            if labels is not None:
                labels = labels[~outlier_indices]
            if predictions is not None:
                predictions = predictions[~outlier_indices]
        if labels is not None:
            edge_colors = np.where(labels.reshape(-1, 1), [positive_data_point_color], [data_point_color])
            face_colors = np.copy(edge_colors)
            face_colors[:, 3] = 0.2
        else:
            edge_colors = [data_point_color]
            face_colors = [(*data_point_color[:3], 0.2)]
        axes.scatter(times, fluxes, c=face_colors, marker='o', edgecolors=edge_colors,
                     linewidths=base_data_point_size / 10, s=base_data_point_size, zorder=3)
        if predictions is not None:
            axes.autoscale(False)
            transparent_prediction_color = (*prediction_color[:3], 0)
            prediction_color_map = LinearSegmentedColormap.from_list('prediction-color-map',
                                                                     [transparent_prediction_color, prediction_color])
            midpoints_between_times = (times[1:] + times[:-1]) / 2
            average_midpoint_distance = np.mean(np.diff(midpoints_between_times))
            extra_start_point = times[0] - average_midpoint_distance
            extra_end_point = times[-1] + average_midpoint_distance
            midpoints_between_times = np.concatenate([[extra_start_point], midpoints_between_times, [extra_end_point]])
            prediction_quad_mesh = axes.pcolormesh(midpoints_between_times, [0, 1], predictions[np.newaxis, :],
                                                   cmap=prediction_color_map, vmin=0, vmax=1)
            transformation = axes.get_xaxis_transform()
            prediction_quad_mesh.set_transform(transformation)
            axes.grid(True)  # Re-enable the grid since pcolormesh disables it.
        if title is not None:
            axes.set_title(title)
        figure.patch.set_alpha(0)  # Transparent figure background while keeping grid background.
        # Need to explicitly state face color or the save will override it.
        axes.set_xlim(x_limits[0], x_limits[1])
        axes.set_ylim(y_limits[0], y_limits[1])
        if save_path is not None:
            plt.savefig(save_path, facecolor=figure.get_facecolor(), dpi=400)
        else:
            plt.show()
        plt.close(figure)


def is_outlier(points: np.ndarray, threshold: float = 5):
    """
    Uses the median absolute deviation to determine if the input data points are "outliers" for the purpose of
    plotting.

    :param points: The observations to search for outliers in.
    :param threshold: The modified z-score to use as a threshold. Observations with a modified z-score based on the
                      median absolute deviation greater than this value will be classified as outliers.
    """
    assert len(points.shape) == 1  # Only designed to work with 1D data.
    median = np.nanmedian(points, axis=0)
    absolute_deviation_from_median = np.abs(points - median)
    median_absolute_deviation_from_median = np.nanmedian(absolute_deviation_from_median)
    modified_z_score = 0.6745 * absolute_deviation_from_median / median_absolute_deviation_from_median
    return modified_z_score > threshold


def create_dual_light_curve_figure(fluxes0, times0, name0, fluxes1, times1, name1, title, x_axis_label='Time (days)',
                                   y_axis_label='Relative flux') -> Figure:
    """
    Plots two light curves together. Mostly for comparing a light curve cleaned by two different methods.

    :param fluxes0: The fluxes of the first plot.
    :param times0: The times of the first plot.
    :param name0: The name of the first plot.
    :param fluxes1: The fluxes of the second plot.
    :param times1: The times of the second plot.
    :param name1: The name of the second plot.
    :param title: The title of the figure.
    :param x_axis_label: The label of the x axis.
    :param y_axis_label: The label of the y axis.
    :return: The resulting figure.
    """
    figure = Figure(title=title, x_axis_label=x_axis_label, y_axis_label=y_axis_label, active_drag='box_zoom')
    add_light_curve(figure, times0, fluxes0, name0, 'firebrick')
    add_light_curve(figure, times1, fluxes1, name1, 'mediumblue')
    return figure


def create_light_curve_figure(fluxes, times, name, title='', x_axis_label='Time (days)',
                              y_axis_label='Relative flux') -> Figure:
    """
    Plots two light curves together. Mostly for comparing a light curve cleaned by two different methods.

    :param fluxes: The fluxes of the plot.
    :param times: The times of the plot.
    :param name: The name of the plot.
    :param title: The title of the figure.
    :param x_axis_label: The label of the x axis.
    :param y_axis_label: The label of the y axis.
    :return: The resulting figure.
    """
    figure = Figure(title=title, x_axis_label=x_axis_label, y_axis_label=y_axis_label, active_drag='box_zoom')
    add_light_curve(figure, times, fluxes, name, 'mediumblue')
    return figure


def add_light_curve(figure, times, fluxes, legend_label, color):
    """Adds a light curve to the figure."""
    fluxes -= np.minimum(np.nanmin(fluxes), 0)
    flux_median = np.median(fluxes)
    figure.line(times, fluxes / flux_median, line_color=color, line_alpha=0.1)
    figure.circle(times, fluxes / flux_median, legend_label=legend_label, line_color=color, line_alpha=0.4,
                  fill_color=color, fill_alpha=0.1)


def add_folded_light_curve(figure, folded_times, fluxes, times):
    """Adds a light curve to the figure."""
    fluxes -= np.minimum(np.nanmin(fluxes), 0)
    flux_median = np.median(fluxes)
    relative_fluxes = fluxes / flux_median
    mapper = LinearColorMapper(palette=Turbo256, low=min(times), high=max(times))
    data_frame = pd.DataFrame({'folded_time': folded_times, 'flux': relative_fluxes, 'time': times})
    color = {'field': 'time', 'transform': mapper}
    figure.circle(source=data_frame, x='folded_time', y='flux', line_color=color, line_alpha=0.4,
                  fill_color=color, fill_alpha=0.1)
    return figure


async def calculate_inlier_range(points: np.ndarray) -> (float, float):
    """
    Calculates the inlier range for a set of points.

    :param points: The points to get the range for.
    :return: The start and end of the inlier range.
    """
    outlier_indices = is_outlier(points)
    inliers = points[~outlier_indices]
    return float(np.nanmin(inliers)), float(np.nanmax(inliers))



"""
Code for a class to represent a light curve. See the contained class docstring for more details.
"""
from __future__ import annotations

from abc import ABC
from typing import Union, List, Optional

import lightkurve.lightcurve
import numpy as np
import pandas as pd
from lightkurve.periodogram import LombScarglePeriodogram
import scipy.signal


class LightCurve(ABC):
    """
    A class to represent a light curve. A light curve is a collection of data which may includes times, fluxes,
    flux errors, and related values.
    """
    def __init__(self):
        self.data_frame: pd.DataFrame = pd.DataFrame()
        self.flux_column_names: List[str] = []
        self.time_column_name: Union[str, None] = None
        self._variability_period: Optional[float] = None
        self._variability_period_epoch: Optional[float] = None
        self.folded_times_column_name = '_folded_times'

    @property
    def fluxes(self) -> np.ndarray:
        """
        The fluxes of the light curve.

        :return: The fluxes.
        """
        return self.data_frame[self.flux_column_names[0]].values

    @fluxes.setter
    def fluxes(self, value: np.ndarray):
        if len(self.flux_column_names) == 0:
            default_flux_name = 'flux'
            self.flux_column_names.append(default_flux_name)
        self.data_frame[self.flux_column_names[0]] = value

    @property
    def times(self) -> np.ndarray:
        """
        The times of the light curve.

        :return: The times.
        """
        return self.data_frame[self.time_column_name].values

    @times.setter
    def times(self, value: np.ndarray):
        if self.time_column_name is None:
            self.time_column_name = 'time'
        self.data_frame[self.time_column_name] = value

    @property
    def folded_times(self):
        if self.folded_times_column_name not in self.data_frame.columns:
            raise MissingFoldedTimes('Light curve has not been folded.')
        return self.data_frame[self.folded_times_column_name].values

    @folded_times.setter
    def folded_times(self, value: np.ndarray):
        self.data_frame[self.folded_times_column_name] = value

    @property
    def variability_period(self) -> float:
        if self._variability_period is None:
            self.get_variability_phase_folding_parameters()
        return self._variability_period

    @property
    def variability_period_epoch(self) -> float:
        if self._variability_period_epoch is None:
            self.get_variability_phase_folding_parameters()
        return self._variability_period_epoch

    def convert_column_to_relative_scale(self, column_name: str):
        """
        Converts a column to relative scale.

        :param column_name: The name of the column to be converted.
        """
        self.data_frame[column_name] = self.data_frame[column_name] / np.nanmedian(self.data_frame[column_name])

    def convert_columns_to_relative_scale(self, column_names: List[str]):
        """
        Converts multiple columns to relative scale.

        :param column_names: The list of names of columns to be converted.
        """
        for column_name in column_names:
            self.convert_column_to_relative_scale(column_name)

    def convert_to_relative_scale(self):
        """
        Converts the light curve to relative scale.
        """
        self.convert_columns_to_relative_scale(self.flux_column_names)

    @classmethod
    def from_times_and_fluxes(cls, times: np.ndarray, fluxes: np.ndarray) -> LightCurve:
        light_curve = LightCurve()
        light_curve.times = times
        light_curve.fluxes = fluxes
        return light_curve

    def to_lightkurve(self) -> lightkurve.lightcurve.LightCurve:
        return lightkurve.lightcurve.LightCurve(time=self.times, flux=self.fluxes)

    def get_variability_phase_folding_parameters(
            self, minimum_period: Optional[float] = None, maximum_period: Optional[float] = None
    ) -> (float, float, float, float, float):
        fold_period, fold_epoch, time_bin_size, minimum_bin_phase, maximum_bin_phase, inlier_lightkurve_light_curve, periodogram, folded_lightkurve_light_curve = self.get_variability_phase_folding_parameters_and_folding_lightkurve_light_curves(minimum_period=minimum_period, maximum_period=maximum_period)
        self._variability_period = fold_period
        self._variability_period_epoch = fold_epoch
        return fold_period, fold_epoch, time_bin_size, minimum_bin_phase, maximum_bin_phase

    def get_variability_phase_folding_parameters_and_folding_lightkurve_light_curves(
            self, minimum_period: Optional[float] = None, maximum_period: Optional[float] = None):
        median_time_step = np.median(np.diff(self.times[~np.isnan(self.times)]))
        lightkurve_light_curve = self.to_lightkurve()
        inlier_lightkurve_light_curve = lightkurve_light_curve.remove_outliers(sigma=3)
        periodogram = LombScarglePeriodogram.from_lightcurve(inlier_lightkurve_light_curve, oversample_factor=100,
                                                             minimum_period=minimum_period,
                                                             maximum_period=maximum_period)
        folded_lightkurve_light_curve = inlier_lightkurve_light_curve.fold(period=periodogram.period_at_max_power,
                                                                           wrap_phase=periodogram.period_at_max_power)
        fold_period = folded_lightkurve_light_curve.period.value
        time_bin_size = fold_period / 25
        binned_folded_lightkurve_light_curve = folded_lightkurve_light_curve.bin(time_bin_size=time_bin_size,
                                                                                 aggregate_func=np.nanmedian)
        minimum_bin_index = np.nanargmin(binned_folded_lightkurve_light_curve.flux.value)
        maximum_bin_index = np.nanargmax(binned_folded_lightkurve_light_curve.flux.value)
        minimum_bin_phase = binned_folded_lightkurve_light_curve.phase.value[minimum_bin_index]
        maximum_bin_phase = binned_folded_lightkurve_light_curve.phase.value[maximum_bin_index]
        fold_epoch = inlier_lightkurve_light_curve.time.value[0]
        return fold_period, fold_epoch, time_bin_size, minimum_bin_phase, maximum_bin_phase, inlier_lightkurve_light_curve, periodogram, folded_lightkurve_light_curve

    def fold(self, period: float, epoch: float) -> None:
        self.folded_times = (self.times - epoch) % period


class MissingFoldedTimes(Exception):
    pass



"""
Code for a database of MOA light curves including non-microlensing, single lensing, and binary lensing collcetions.
"""
from ramjet.data_interface.moa_data_interface import MoaDataInterface
from ramjet.photometric_database.derived.moa_survey_light_curve_collection import MoaSurveyLightCurveCollection
from ramjet.photometric_database.standard_and_injected_light_curve_database import \
    StandardAndInjectedLightCurveDatabase
from ramjet.photometric_database.light_curve_dataset_manipulations import OutOfBoundsInjectionHandlingMethod, \
    BaselineFluxEstimationMethod


class MoaSurveyNoneSingleAndBinaryDatabase(StandardAndInjectedLightCurveDatabase):
    """
    A class for a database of MOA light curves including non-microlensing, single lensing, and binary lensing
    collections.
    """
    moa_data_interface = MoaDataInterface()

    def __init__(self):
        super().__init__()
        self.number_of_label_values = 2
        self.number_of_parallel_processes_per_map = 5
        self.time_steps_per_example = 18000
        self.out_of_bounds_injection_handling = OutOfBoundsInjectionHandlingMethod.RANDOM_INJECTION_LOCATION
        self.baseline_flux_estimation_method = BaselineFluxEstimationMethod.MEDIAN_ABSOLUTE_DEVIATION
        self.shuffle_buffer_size = 1000
        # self.include_time_as_channel = True
        # self.include_flux_errors_as_channel = True
        negative_training = MoaSurveyLightCurveCollection(
            survey_tags=['v', 'n', 'nr', 'm', 'j', self.moa_data_interface.no_tag_string], label=[0, 0],
            dataset_splits=list(range(8)))
        self.training_standard_light_curve_collections = [
            negative_training,
            MoaSurveyLightCurveCollection(survey_tags=['c', 'cf', 'cp', 'cw', 'cs'], label=[1, 0],
                                          dataset_splits=list(range(8))),
            MoaSurveyLightCurveCollection(survey_tags=['cb'], label=[1, 1],
                                          dataset_splits=list(range(8)))
        ]
        self.validation_standard_light_curve_collections = [
            MoaSurveyLightCurveCollection(survey_tags=['v', 'n', 'nr', 'm', 'j', self.moa_data_interface.no_tag_string],
                                          label=[0, 0], dataset_splits=[8]),
            MoaSurveyLightCurveCollection(survey_tags=['c', 'cf', 'cp', 'cw', 'cs'], label=[1, 0],
                                          dataset_splits=[8]),
            MoaSurveyLightCurveCollection(survey_tags=['cb'], label=[1, 1],
                                          dataset_splits=[8])
        ]
        self.inference_light_curve_collections = [
            MoaSurveyLightCurveCollection(survey_tags=['v', 'n', 'nr', 'm', 'j', self.moa_data_interface.no_tag_string],
                                          label=[0, 0], dataset_splits=[9]),
            MoaSurveyLightCurveCollection(survey_tags=['c', 'cf', 'cp', 'cw', 'cs'], label=[1, 0],
                                          dataset_splits=[9]),
            MoaSurveyLightCurveCollection(survey_tags=['cb'], label=[1, 1],
                                          dataset_splits=[9])
        ]



from dataclasses import dataclass
from functools import partial
from typing import List, Callable, Any, Self

import numpy as np
from torch.utils.data import Dataset

from qusi.light_curve_collection import LabeledLightCurveCollection
from qusi.light_curve_dataset import default_light_curve_observation_post_injection_transform


@dataclass
class FiniteStandardLightCurveObservationDataset(Dataset):
    standard_light_curve_collections: List[LabeledLightCurveCollection]
    post_injection_transform: Callable[[Any], Any]
    length: int
    collection_start_indexes: List[int]

    @classmethod
    def new(cls, standard_light_curve_collections: List[LabeledLightCurveCollection]) -> Self:
        length = 0
        collection_start_indexes: List[int] = []
        for standard_light_curve_collection in standard_light_curve_collections:
            standard_light_curve_collection_length = len(
                list(standard_light_curve_collection.path_getter.get_paths()))
            collection_start_indexes.append(length)
            length += standard_light_curve_collection_length
        instance = cls(standard_light_curve_collections=standard_light_curve_collections,
                       post_injection_transform=partial(default_light_curve_observation_post_injection_transform,
                                                        length=2500),
                       length=length,
                       collection_start_indexes=collection_start_indexes)
        return instance

    def __len__(self) -> int:
        return self.length

    def __getitem__(self, index) -> Any:
        collection_index = np.searchsorted(self.collection_start_indexes, index, side='right') - 1
        collection = self.standard_light_curve_collections[collection_index]
        index_in_collection = index - self.collection_start_indexes[collection_index]
        observation = collection[index_in_collection]
        transformed_observation = self.post_injection_transform(observation)
        return transformed_observation



from typing import List

import torch
from torch.nn import Module
from torch.types import Device
from torch.utils.data import DataLoader

from qusi.finite_standard_light_curve_observation_dataset import FiniteStandardLightCurveObservationDataset


def finite_datasets_test_session(test_datasets: List[FiniteStandardLightCurveObservationDataset], model: Module,
                                 metric_functions: List[Module], batch_size: int, device: Device):
    test_dataloaders: List[DataLoader] = []
    for test_dataset in test_datasets:
        test_dataloaders.append(DataLoader(test_dataset, batch_size=batch_size, pin_memory=True))
    model.eval()
    results = []
    for test_dataloader in test_dataloaders:
        result = finite_dataset_test_phase(test_dataloader, model, metric_functions, device=device)
        results.append(result)
    return results


def get_device():
    if torch.cuda.is_available():
        device = torch.device('cuda')
    else:
        device = torch.device('cpu')
    return device


def finite_dataset_test_phase(dataloader, model: Module, metric_functions: List[Module], device: Device):
    batch_count = 0
    metric_totals = torch.zeros(size=[len(metric_functions)])
    model.eval()
    with torch.no_grad():
        for input_features, targets in dataloader:
            input_features = input_features.to(device, non_blocking=True)
            targets = targets.to(device, non_blocking=True)
            predicted_targets = model(input_features)
            for metric_function_index, metric_function in enumerate(metric_functions):
                batch_metric_value = metric_function(predicted_targets.to(device, non_blocking=True),
                                                     targets)
                metric_totals[metric_function_index] += batch_metric_value.to('cpu', non_blocking=True)
            batch_count += 1
    cycle_metric_values = metric_totals / batch_count
    return cycle_metric_values



from dataclasses import dataclass


@dataclass
class TrainSystemConfiguration:
    """
    Configuration settings for the system of a train session.

    :ivar preprocessing_processes_per_train_process: The number of processes that are started to preprocess the data
        per train process. The train session will create this many processes for each the train data and the validation
        data.
    """
    preprocessing_processes_per_train_process: int

    @classmethod
    def new(cls, preprocessing_processes_per_train_process: int = 10):
        return cls(preprocessing_processes_per_train_process=preprocessing_processes_per_train_process)



from pathlib import Path

import pandas as pd
from astropy import units
from astropy.coordinates import SkyCoord
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

from ramjet.photometric_database.tess_ffi_light_curve import TessFfiLightCurve


def create_pdf_of_light_curve_variability(light_curve_path: Path, output_pdf_path: Path) -> None:
    light_curve = TessFfiLightCurve.from_path(light_curve_path)
    fold_period, fold_epoch, time_bin_size, minimum_bin_phase, maximum_bin_phase, inlier_lightkurve_light_curve, periodogram, folded_lightkurve_light_curve = light_curve.get_variability_phase_folding_parameters_and_folding_lightkurve_light_curves()
    variability_centroid_and_frames = \
        light_curve.estimate_photometric_variability_centroid_and_frames_from_ffi_based_on_folding_parameters(fold_epoch,
                                                                                                              fold_period,
                                                                                                              maximum_bin_phase,
                                                                                                              minimum_bin_phase,
                                                                                                              time_bin_size)
    centroid_sky_coord = variability_centroid_and_frames[0]
    target_pixel_file = variability_centroid_and_frames[1]
    difference_target_pixel_frame = variability_centroid_and_frames[2]
    median_maximum_target_pixel_frame = variability_centroid_and_frames[3]
    median_minimum_target_pixel_frame = variability_centroid_and_frames[4]
    with PdfPages(output_pdf_path) as pdf_file:
        inlier_lightkurve_light_curve_figure, inlier_lightkurve_light_curve_axes = plt.subplots()
        inlier_lightkurve_light_curve.scatter(ax=inlier_lightkurve_light_curve_axes)
        pdf_file.savefig(inlier_lightkurve_light_curve_figure)
        periodogram_figure, periodogram_axes = plt.subplots()
        periodogram.plot(ax=periodogram_axes)
        pdf_file.savefig(periodogram_figure)
        folded_lightkurve_light_curve_figure, folded_lightkurve_light_curve_axes = plt.subplots()
        folded_lightkurve_light_curve.scatter(ax=folded_lightkurve_light_curve_axes)
        pdf_file.savefig(folded_lightkurve_light_curve_figure)
        median_minimum_target_pixel_frame_figure, median_minimum_target_pixel_frame_axes = plt.subplots()
        median_minimum_target_pixel_frame.plot(ax=median_minimum_target_pixel_frame_axes)
        pdf_file.savefig(median_minimum_target_pixel_frame_figure)
        median_maximum_target_pixel_frame_figure, median_maximum_target_pixel_frame_axes = plt.subplots()
        median_maximum_target_pixel_frame.plot(ax=median_maximum_target_pixel_frame_axes)
        pdf_file.savefig(median_maximum_target_pixel_frame_figure)
        difference_target_pixel_frame_figure, difference_target_pixel_frame_axes = plt.subplots()
        difference_target_pixel_frame.plot(ax=difference_target_pixel_frame_axes)
        centroid_pixel_position = target_pixel_file.wcs.world_to_pixel(centroid_sky_coord)
        difference_target_pixel_frame_axes.plot(target_pixel_file.column + centroid_pixel_position[0],
                 target_pixel_file.row + centroid_pixel_position[1], 'co')
        difference_target_pixel_frame_axes.plot(target_pixel_file.column + centroid_pixel_position[0],
                 target_pixel_file.row + centroid_pixel_position[1], 'mo')
        target_sky_coord = SkyCoord(ra=target_pixel_file.ra, dec=target_pixel_file.dec, unit=units.deg)
        target_pixel_position = target_pixel_file.wcs.world_to_pixel(target_sky_coord)
        difference_target_pixel_frame_axes.plot(target_pixel_position[0] + target_pixel_file.column,
                 target_pixel_position[1] + target_pixel_file.row, 'ro')
        pdf_file.savefig(difference_target_pixel_frame_figure)
        plt.close('all')  # TODO: This is a hack. Should really use proper figure contexts or something.


def create_pdf_of_variability_of_filtered_light_curves_csv(csv_path: Path, output_directory_path: Path):
    filtered_data_frame = pd.read_csv(csv_path)
    # filtered_data_frame = filtered_data_frame.tail(10)
    output_directory_path.mkdir(exist_ok=True, parents=True)
    skipped = []
    for index, row in filtered_data_frame.iterrows():
        light_curve_path = Path(row['light_curve_path'])
        tic_id, sector = row['tic_id'], row['sector']
        pdf_path = output_directory_path.joinpath(f'{index}_tic_id_{tic_id}_sector_{sector}')
        print('=' * 50)
        print(pdf_path)
        try:
            create_pdf_of_light_curve_variability(light_curve_path, pdf_path)
        except:
            skipped.append(index)
        print('=' * 50)
    print(f'Skipped: {len(skipped)}')
    print(f'{skipped}')


if __name__ == '__main__':
    filtered_csv_path_ = Path('/att/gpfsfs/briskfs01/ppl/golmsche/generalized-photometric-neural-network-experiments/logs/FfiHades_corrected_non_rrl_label_no_bn_2022_02_03_23_52_12/filtered_infer_results_2022-02-06-13-21-41.csv')
    output_directory_path_ = filtered_csv_path_.parent.joinpath('variability_pdfs')
    create_pdf_of_variability_of_filtered_light_curves_csv(filtered_csv_path_, output_directory_path_)



"""
Code for a light curve collection that stores its metadata in the SQL database.
"""
import random
from pathlib import Path
from typing import Iterable, List, Union
from uuid import uuid4

from peewee import Select, Field, Case

from ramjet.data_interface.metadatabase import MetadatabaseModel
from ramjet.photometric_database.light_curve_collection import LightCurveCollection, \
    LightCurveCollectionMethodNotImplementedError


class SqlMetadataLightCurveCollection(LightCurveCollection):
    """
    Class for a light curve collection that stores its metadata in the SQL database.
    """
    def __init__(self):
        super().__init__()

    def get_sql_query(self) -> Select:
        """
        Gets the SQL query for the database models for the light curve collection.

        :return: The SQL query.
        """
        raise LightCurveCollectionMethodNotImplementedError

    def sql_count(self) -> int:
        """
        Gets the count of the rows returned by the SQL query.

        :return: The count.
        """
        return self.get_sql_query().count()

    def get_path_from_model(self, model: MetadatabaseModel) -> Path:
        """
        Gets the light curve path from the SQL database model.

        :return: The path to the light curve.
        """
        raise LightCurveCollectionMethodNotImplementedError

    def get_paths(self) -> Iterable[Path]:
        """
        Gets the paths for the light curves in the collection.

        :return: An iterable of the light curve paths.
        """
        query = self.get_sql_query()
        query = query.objects().iterator()  # Disable Peewee's cache and graph for better performance.
        for model in query:
            yield Path(self.get_path_from_model(model))

    @staticmethod
    def order_by_uuid_with_random_start(select_query: Select, uuid_field: Field) -> Select:
        """
        Applies an "order by" on a query using a passed UUID field. The "order by" starts at a random UUID and then
        loops back to the minimum UUID include all entities.

        :param select_query: The query to add the "order by" to.
        :param uuid_field: The UUID field to order on.
        :return: The query updated to include the "order by".
        """
        random_start_case = Case(None, [(uuid_field > uuid4(), 0)], 1)
        updated_select_query = select_query.order_by(random_start_case, uuid_field)
        return updated_select_query

    @staticmethod
    def order_by_dataset_split_with_random_start(select_query: Select, dataset_split_field: Field,
                                                 available_dataset_splits: Union[List[int], None]) -> Select:
        """
        Applies an "order by" on a query using a passed dataset_split field. The "order by" starts at a random
        dataset_split out of the passed available options, then loops back to the minimum dataset_split to include all
        entities.

        :param select_query: The query to add the "order by" to.
        :param dataset_split_field: The dataset_split field to order on.
        :param available_dataset_splits: The available dataset_splits to start on.
        :return: The query updated to include the "order by".
        """
        if available_dataset_splits is None:
            available_dataset_splits = list(range(10))
        start_dataset_split = random.choice(available_dataset_splits)
        start_case = Case(None, [(dataset_split_field >= start_dataset_split, 0)], 1)
        updated_select_query = select_query.order_by(start_case, dataset_split_field)
        return updated_select_query



import numpy as np

from ramjet.photometric_database.derived.toy_light_curve_collection import ToyFlatLightCurveCollection, \
    ToySineWaveLightCurveCollection, ToyFlatAtValueLightCurveCollection
from ramjet.photometric_database.standard_and_injected_light_curve_database import StandardAndInjectedLightCurveDatabase


class ToyRamjetDatabase(StandardAndInjectedLightCurveDatabase):
    def __init__(self):
        super().__init__()
        self.batch_size = 10
        self.number_of_parallel_processes_per_map = 1
        self.time_steps_per_example = 100
        self.training_standard_light_curve_collections = [
            ToyFlatLightCurveCollection(),
            ToySineWaveLightCurveCollection(),
        ]
        self.validation_standard_light_curve_collections = [
            ToyFlatLightCurveCollection(),
            ToySineWaveLightCurveCollection(),
        ]
        self.inference_light_curve_collections = [
            ToyFlatLightCurveCollection(),
            ToySineWaveLightCurveCollection(),
        ]

class ToyRamjetDatabaseWithAuxiliary(StandardAndInjectedLightCurveDatabase):
    def __init__(self):
        super().__init__()
        self.batch_size = 10
        self.number_of_parallel_processes_per_map = 1
        self.time_steps_per_example = 100
        self.number_of_auxiliary_values = 2
        flat_collection = ToyFlatLightCurveCollection()
        sine_wave_collection = ToySineWaveLightCurveCollection()
        flat_collection.load_auxiliary_information_for_path = lambda path: np.array([0, 0], dtype=np.float32)
        sine_wave_collection.load_auxiliary_information_for_path = lambda path: np.array([1, 1], dtype=np.float32)
        self.training_standard_light_curve_collections = [
            flat_collection,
            sine_wave_collection,
        ]
        self.validation_standard_light_curve_collections = [
            flat_collection,
            sine_wave_collection,
        ]
        self.inference_light_curve_collections = [
            flat_collection,
            sine_wave_collection,
        ]


class ToyRamjetDatabaseWithFlatValueAsLabel(StandardAndInjectedLightCurveDatabase):
    def __init__(self):
        super().__init__()
        self.batch_size = 10
        self.number_of_parallel_processes_per_map = 1
        self.time_steps_per_example = 100
        self.normalize_fluxes = lambda fluxes: fluxes
        self.training_standard_light_curve_collections = [
            ToyFlatAtValueLightCurveCollection(),
        ]
        self.validation_standard_light_curve_collections = [
            ToyFlatAtValueLightCurveCollection(),
        ]
        self.inference_light_curve_collections = [
            ToyFlatAtValueLightCurveCollection(),
        ]


"""Script that generates a Gravitational Microlensing Signal, randomly, within the natural parameters: u0 (
source-lens impact parameter), tE (Einstein radius crossing time), rho (angular source size normalized by the
angular Einstein radius) , s (Projected separation of the masses normalized by the angular Einstein radius),
q (Mass ratio M_planet/M_host), alpha (Trajectory angle). The distribution for tE and rho are based on the MOA
observations.
"""
import requests
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pathlib import Path
try:
    from muLAn.models.vbb.vbb import vbbmagU
except ModuleNotFoundError as error:
    vbbmagU = None


class MagnificationSignal:
    """A class to generate a random microlensing magnification signal.
    Using the parameters:
    u0 (source-lens impact parameter)
    tE (Einstein radius crossing time)
    rho (angular source size normalized by the angular Einstein radius)
    s (Projected separation of the masses normalized by the angular Einstein radius)
    q (Mass ratio M_planet/M_host)
    alpha (Trajectory angle)
    > The distribution for tE and rho are based on the MOA observations
    > No parallax effect is considered
    """
    tE_list: np.ndarray = None
    rho_list: np.ndarray = None

    def __init__(self):
        self.load_moa_meta_data_to_class_attributes()
        self.n_data_points = 80000
        self.timeseries = np.linspace(-30, 30, self.n_data_points)
        self.magnification = None
        self.magnification_signal_curve = None
        self.u0 = None
        self.tE = None
        self.rho = None
        self.s = None
        self.q = None
        self.alpha = None

    def load_moa_meta_data_to_class_attributes(self):
        """
        Loads the MOA meta data defining microlensing to class attributes. If already loaded, does nothing.
        """
        if self.tE_list is None:
            microlensing_meta_data_path = Path(__file__).parent.joinpath(
                'microlensing_signal_meta_data/candlist_RADec.dat.txt')
            microlensing_meta_data_path.parent.mkdir(parents=True, exist_ok=True)
            if not microlensing_meta_data_path.exists():
                candidate_list_csv_url = 'https://exoplanetarchive.ipac.caltech.edu/data/ExoData/MOA/candlist_RADec.dat'
                response = requests.get(candidate_list_csv_url)
                with open(microlensing_meta_data_path, 'wb') as csv_file:
                    csv_file.write(response.content)
            data = pd.read_csv(microlensing_meta_data_path, header=None, delim_whitespace=True, comment='#',
                               usecols=[19, 36], names=['tE', 'rho'])
            self.tE_list: np.ndarray = data['tE'].values
            self.rho_list: np.ndarray = data['rho'].values
            bad_indexes = np.argwhere(self.tE_list > 6000)
            self.tE_list = np.delete(self.tE_list, bad_indexes)
            self.rho_list = np.delete(self.rho_list, bad_indexes)

    def getting_random_values(self):
        """
        Set randomly the natural parameters: u0 (source-lens impact parameter), tE (Einstein radius crossing time),
        rho (angular source size normalized by the angular Einstein radius) , s (Projected separation of the masses
        normalized by the angular Einstein radius), q (Mass ratio M_planet/M_host), alpha (Trajectory angle)
        """

        u0_list = np.linspace(-0.1, 0, 1000)
        self.u0 = np.random.choice(u0_list)

        index = np.random.choice(np.arange(self.tE_list.shape[0]))
        self.tE = float(self.tE_list[index])
        self.rho = float(self.rho_list[index])

        s_list = np.linspace(0.7, 1.3, 100)
        self.s = np.random.choice(s_list)

        q_list = np.power(10, (np.linspace(-2.5, -0.3, 1000)))
        self.q = np.random.choice(q_list)

        pi_denomintor = 128
        alpha_list = np.concatenate([np.linspace(0, np.pi / pi_denomintor),
                                     np.linspace(np.pi - (np.pi / pi_denomintor), np.pi)])
        self.alpha = np.random.choice(alpha_list)

    def generating_magnification(self):
        """
        Creates the magnification signal
        """
        lens_params = dict({'u0': self.u0,
                            'tE': self.tE,
                            't0': 0.0,
                            'rho': self.rho,
                            's': self.s,
                            'q': self.q,
                            'alpha': self.alpha
                            })

        # Compute magnification
        self.magnification = self.calculating_magnification_from_vbb(self.timeseries, lens_params)
        self.magnification_signal_curve = pd.DataFrame({'Time': self.timeseries, 'Magnification': self.magnification})

    def plot_magnification(self):
        """
        Plot the light curve.
        """
        # PLOT
        plt.plot(self.timeseries, self.magnification)
        plt.xlabel('Days')
        plt.ylabel('Magnification')
        plt.title(f'u0= {self.u0:3.5f}; tE= {self.tE:12.5f}; rho= {self.rho:8.5f};\n s= {self.s:3.5f}; '
                  f'q= {self.q:8.5f}; alpha= {self.alpha:3.5f}')
        # plt.title(f'u0= {:3.5f}; tE= {:10.5f}; rho= {:5.5f};\n s= {:3.5f}; '
        #           f'q= {:5.5f}; alpha= {:2.5f}'.format(self.u0,self.tE,self.rho,self.s,self.q,self.alpha))
        plt.show()

    @classmethod
    def generate_randomly_based_on_moa_observations(cls, time_range: float = 30):
        microlensing_signal = cls()
        microlensing_signal.timeseries = np.linspace(-time_range, time_range, microlensing_signal.n_data_points)
        microlensing_signal.getting_random_values()
        microlensing_signal.generating_magnification()
        return microlensing_signal

    @classmethod
    def generate_approximately_pspl_randomly_based_on_moa_observations(cls, time_range: float = 30):
        microlensing_signal = cls()
        microlensing_signal.timeseries = np.linspace(-time_range, time_range, microlensing_signal.n_data_points)
        microlensing_signal.getting_random_values()
        microlensing_signal.q = 0.00001
        microlensing_signal.s = 10
        microlensing_signal.alpha = np.pi / 2
        microlensing_signal.generating_magnification()
        return microlensing_signal

    @staticmethod
    def calculating_magnification_from_vbb(timeseries, lens_params):
        """Return the VBB method finite-source uniform magnification.
        Adapted from muLAn: gravitational MICROlensing Analysis Software.
        """
        # Get parameters
        t0 = lens_params['t0']
        u0 = lens_params['u0']
        tE = lens_params['tE']
        rho = lens_params['rho']
        q = lens_params['q']
        alpha = lens_params['alpha']
        s = lens_params['s']

        tau = (timeseries - t0) / tE

        cos_alpha = np.cos(alpha)
        sin_alpha = np.sin(alpha)

        x = tau * cos_alpha - u0 * sin_alpha
        y = tau * sin_alpha + u0 * cos_alpha

        # Conversion secondary body left -> right
        x = -x
        # Compute magnification
        accuracy = 1.e-3  # Absolute mag accuracy (mag+/-accuracy)
        magnification = np.array([vbbmagU(s, q, rho, x[i], y[i], accuracy) for i in range(len(x))])
        return magnification


if __name__ == '__main__':
    import time
    start_time = time.time()
    random_signal = MagnificationSignal.generate_randomly_based_on_moa_observations()
    print("--- %s seconds ---" % (time.time() - start_time))
    random_signal.plot_magnification()
    print("Done")



from ramjet.data_interface.tess_data_interface import TessFluxType
from ramjet.photometric_database.derived.self_lensing_binary_synthetic_signals_light_curve_collection import \
    SelfLensingBinarySyntheticSignalsLightCurveCollection, ReversedSelfLensingBinarySyntheticSignalsLightCurveCollection
from ramjet.photometric_database.derived.siddhant_solanki_heart_beat_synthetic_signals_collection import \
    SiddhantSolankiHeartBeatSyntheticSignalsCollection, SiddhantSolankiNonHeartBeatSyntheticSignalsCollection, \
    TessFfiHeartBeatHardNegativeLightcurveCollection
from ramjet.photometric_database.derived.tess_ffi_light_curve_collection import TessFfiLightCurveCollection
from ramjet.photometric_database.derived.tess_ffi_transit_databases import TessFfiDatabase
from ramjet.photometric_database.derived.tess_two_minute_cadence_light_curve_collection import TessTwoMinuteCadenceLightCurveCollection
from ramjet.photometric_database.standard_and_injected_light_curve_database import StandardAndInjectedLightCurveDatabase


magnitude_range = (0, 15)


class SiddhantSolankiHeartBeatSyntheticDatabase(TessFfiDatabase):
    def __init__(self):
        super().__init__()
        self.training_standard_light_curve_collections = [
            TessFfiLightCurveCollection(magnitude_range=magnitude_range),
            TessFfiHeartBeatHardNegativeLightcurveCollection(magnitude_range=magnitude_range)]
        self.training_injectee_light_curve_collection = TessFfiLightCurveCollection(magnitude_range=magnitude_range)
        self.training_injectable_light_curve_collections = [
            SiddhantSolankiHeartBeatSyntheticSignalsCollection(),
            SiddhantSolankiNonHeartBeatSyntheticSignalsCollection(),
        ]
        self.validation_standard_light_curve_collections = [
            TessFfiLightCurveCollection(magnitude_range=magnitude_range)
        ]
        self.validation_injectee_light_curve_collection = self.training_injectee_light_curve_collection
        self.validation_injectable_light_curve_collections = self.training_injectable_light_curve_collections
        self.inference_light_curve_collections = [TessFfiLightCurveCollection(magnitude_range=magnitude_range)]


import re
import shutil

import numpy as np
import pandas as pd
from pathlib import Path
from typing import List, Union, Iterable
import socket
import scipy.stats

from filelock import FileLock

from ramjet.data_interface.moa_data_interface import MoaDataInterface
from ramjet.photometric_database.light_curve_collection import LightCurveCollection


class MoaSurveyLightCurveCollection(LightCurveCollection):
    """
    A collection of light curves based on the MOA 9-year survey.
    """
    moa_data_interface = MoaDataInterface()

    def __init__(self, survey_tags: List[str], dataset_splits: Union[List[int], None] = None,
                 label: Union[float, List[float], np.ndarray, None] = None):
        super().__init__()
        self.label = label
        self.survey_tags: List[str] = survey_tags
        self.dataset_splits: Union[List[int], None] = dataset_splits

    def get_paths(self) -> Iterable[Path]:
        """
        Gets the paths for the light curves in the collection.

        :return: An iterable of the light curve paths.
        """
        paths: List[Path] = []
        for tag in self.survey_tags:
            tag_paths = self.moa_data_interface.survey_tag_to_path_list_dictionary[tag]
            if self.dataset_splits is not None:
                # Split on each tag, so that the splitting remains across collections with different tag selections.
                tag_paths = self.shuffle_and_split_paths(tag_paths, self.dataset_splits)
            paths.extend(tag_paths)
        return paths

    def move_path_to_nvme(self, path: Path) -> Path:
        match = re.match(r"gpu\d{3}", socket.gethostname())
        if match is not None:
            nvme_path = Path("/lscratch/golmsche").joinpath(path)
            if not nvme_path.exists():
                nvme_path.parent.mkdir(exist_ok=True, parents=True)
                nvme_lock_path = nvme_path.parent.joinpath(nvme_path.name + '.lock')
                lock = FileLock(str(nvme_lock_path))
                with lock.acquire():
                    if not nvme_path.exists():
                        nvme_tmp_path = nvme_path.parent.joinpath(nvme_path.name + '.tmp')
                        shutil.copy(path, nvme_tmp_path)
                        nvme_tmp_path.rename(nvme_path)
            return nvme_path
        else:
            return path


    def load_times_and_fluxes_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and fluxes from a given light curve path.

        :param path: The path to the light curve file.
        :return: The times and the fluxes of the light curve.
        """
        path = self.move_path_to_nvme(path)
        light_curve_dataframe = pd.read_feather(path)
        times = light_curve_dataframe['HJD'].values
        fluxes = light_curve_dataframe['flux'].values
        return times, fluxes

    def load_times_and_magnifications_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and magnifications from a given path as an injectable signal.

        :param path: The path to the light curve/signal file.
        :return: The times and the magnifications of the light curve/signal.
        """
        path = self.move_path_to_nvme(path)
        times, fluxes = self.load_times_and_fluxes_from_path(path)
        magnifications, times = self.generate_synthetic_signal_from_real_data(fluxes, times)
        return times, magnifications

    @staticmethod
    def generate_synthetic_signal_from_real_data(fluxes: np.ndarray, times: np.ndarray) -> (np.ndarray, np.ndarray):
        """
        Takes real light curve data and converts it to a form that can be used for synthetic light curve injection.

        :param fluxes: The real light curve fluxes.
        :param times: The real light curve times.
        :return: Fake synthetic magnifications and times.
        """
        flux_median_absolute_deviation = scipy.stats.median_abs_deviation(fluxes)
        normalized_fluxes = (fluxes / flux_median_absolute_deviation) * 0.25
        # relative_times = times - np.min(times)
        return normalized_fluxes, times


"""
Code for creating database based on the MOA microlensing data.
"""
from ramjet.photometric_database.derived.moa_microlensing_light_curve_collection import \
    MOAPositiveMicrolensingLightCurveCollection, MOANegativeMicrolensingLightCurveCollection, \
    MicrolensingSyntheticGeneratedDuringRunningSignalCollection
from ramjet.photometric_database.standard_and_injected_light_curve_database import StandardAndInjectedLightCurveDatabase


class MoaMicrolensingDatabase(StandardAndInjectedLightCurveDatabase):
    """
    A database to train a network to find microlensing events in MOA data. Uses real data that were previous classified.
    """

    def __init__(self):
        super().__init__()
        self.training_standard_light_curve_collections = [
            MOAPositiveMicrolensingLightCurveCollection(dataset_splits=[0, 2, 3, 4], split_pieces=5),
            MOANegativeMicrolensingLightCurveCollection(dataset_splits=[0, 2, 3, 4], split_pieces=5)
        ]
        self.validation_standard_light_curve_collections = [
            MOAPositiveMicrolensingLightCurveCollection(dataset_splits=[1], split_pieces=5),
            MOANegativeMicrolensingLightCurveCollection(dataset_splits=[1], split_pieces=5)
        ]
        self.inference_light_curve_collections = MOANegativeMicrolensingLightCurveCollection(dataset_splits=[1],
                                                                                             split_pieces=5)


class MoaMicrolensingWithSyntheticDatabase(StandardAndInjectedLightCurveDatabase):
    """
    A database to train a network to find microlensing events in MOA data. Uses synthetic data injected on previous
    data classified as negative.
    """

    def __init__(self):
        super().__init__()
        self.allow_out_of_bounds_injection = True
        self.training_standard_light_curve_collections = [
            MOAPositiveMicrolensingLightCurveCollection(),
            MOANegativeMicrolensingLightCurveCollection()
        ]
        self.training_injectee_light_curve_collection = MOANegativeMicrolensingLightCurveCollection()
        self.training_injectable_light_curve_collections = MicrolensingSyntheticGeneratedDuringRunningSignalCollection()
        self.validation_standard_light_curve_collections = self.training_standard_light_curve_collections



"""
Code to for a class to represent a TESS two minute cadence light curve.
"""
from __future__ import annotations

import re
import sys

import numpy as np
from enum import Enum
from pathlib import Path
from typing import List, Union, Any, Optional
from astropy.io import fits

from ramjet.data_interface.tess_data_interface import download_two_minute_cadence_light_curve
from ramjet.photometric_database.tess_light_curve import TessLightCurve


class TessMissionLightCurveColumnName(Enum):
    """
    An enum to represent the column names of the TESS two minute cadence data.
    """
    TIME__BTJD = 'time__btjd'
    SAP_FLUX = 'sap_flux'
    PDCSAP_FLUX = 'pdcsap_flux'
    SAP_FLUX_ERROR = 'sap_flux_error'
    PDCSAP_FLUX_ERROR = 'pdcsap_flux_error'


class TessMissionLightCurveFitsIndex(Enum):
    """
    An enum to represent the indexes of the TESS two minute cadence data in MAST FITS files.
    """
    TIME__BTJD = 'TIME'
    SAP_FLUX = 'SAP_FLUX'
    PDCSAP_FLUX = 'PDCSAP_FLUX'
    SAP_FLUX_ERROR = 'SAP_FLUX_ERR'
    PDCSAP_FLUX_ERROR = 'PDCSAP_FLUX_ERR'


class TessMissionLightCurve(TessLightCurve):
    """
    A class to represent a TESS two minute cadence light curve.
    """

    def __init__(self):
        super().__init__()
        self.flux_column_names = [TessMissionLightCurveColumnName.PDCSAP_FLUX.value,
                                  TessMissionLightCurveColumnName.SAP_FLUX.value]

    @classmethod
    def from_path(cls, path: Path, fits_indexes_to_load: Union[List[TessMissionLightCurveFitsIndex], None] = None
                  ) -> TessMissionLightCurve:
        """
        Creates a TESS two minute light curve from a path to the MAST FITS file.

        :param path: The path to the FITS file to load.
        :param fits_indexes_to_load: The indexes to load from the FITS file. By default, all will be loaded. Selecting
                                     specific ones may speed the process when loading many light curves.
        :return: The light curve.
        """
        light_curve = cls()
        light_curve.time_column_name = TessMissionLightCurveColumnName.TIME__BTJD.value
        if fits_indexes_to_load is None:
            fits_indexes_to_load = list(TessMissionLightCurveFitsIndex)
        with fits.open(path) as hdu_list:
            light_curve_table = hdu_list[1].data  # Light curve information is in first extension table.
            for fits_index in fits_indexes_to_load:
                column_name = TessMissionLightCurveColumnName[fits_index.name]
                column = light_curve_table[fits_index.value]
                column = ensure_native_byte_order(column)
                light_curve.data_frame[column_name.value] = column
        light_curve.tic_id, light_curve.sector = cls.get_tic_id_and_sector_from_file_path(path)
        return light_curve

    @classmethod
    def from_mast(cls, tic_id: int, sector: Optional[int] = None,
                  fits_indexes_to_load: Union[List[TessMissionLightCurveFitsIndex], None] = None
                  ) -> TessMissionLightCurve:
        """
        Downloads a FITS file from MAST and creates a TESS two minute light curve from it.

        :param tic_id: The TIC ID of the target.
        :param sector: The sector of the observation.
        :param fits_indexes_to_load: The indexes to load from the FITS file. By default, all will be loaded. Selecting
                                     specific ones may speed the process when loading many light curves.
        :return: The light curve.
        """
        light_curve_path = download_two_minute_cadence_light_curve(tic_id=tic_id, sector=sector)
        light_curve = cls.from_path(path=light_curve_path, fits_indexes_to_load=fits_indexes_to_load)
        return light_curve

    @classmethod
    def from_identifier(cls, identifier: Any) -> TessMissionLightCurve:
        """
        Loads the light curve in a generalized way, attempting to infer the light curve based on the passed identifier.

        :param identifier: The identifier of the light curve. Could come in various forms.
        :return: The light curve.
        """
        integer_types = (int, np.integer)
        if isinstance(identifier, Path):
            return cls.from_path(path=identifier)
        elif isinstance(identifier, tuple) and (isinstance(identifier[0], integer_types) and
                                                isinstance(identifier[1], integer_types)):
            tic_id = identifier[0]
            sector = identifier[1]
            return cls.from_mast(tic_id=tic_id, sector=sector)
        elif isinstance(identifier, str):
            tic_id, sector = cls.get_tic_id_and_sector_from_identifier_string(identifier)
            return cls.from_mast(tic_id=tic_id, sector=sector)
        else:
            raise ValueError(f'{identifier} does not match a known type to infer the light curve identifier from.')

    @staticmethod
    def get_tic_id_and_sector_from_file_path(file_path: Path) -> (int, Union[int, None]):
        """
        Gets the TIC ID and sector from commonly encountered file name patterns.

        :param file_path: The path of the file to extract the TIC ID and sector.
        :return: The TIC ID and sector. The sector might be omitted (as None).
        """
        file_name = file_path.stem
        tic_id, sector = TessMissionLightCurve.get_tic_id_and_sector_from_identifier_string(file_name)
        return tic_id, sector

    @staticmethod
    def get_tic_id_and_sector_from_identifier_string(identifier_string: str) -> (int, Union[int, None]):
        """
        Gets the TIC ID and sector from commonly encountered identifier string patterns.

        :param identifier_string: The string to extract the TIC ID and sector.
        :return: The TIC ID and sector. The sector might be omitted (as None).
        """
        # Search for the human-readable version. E.g., "TIC 169480782 sector 5"
        match = re.search(r'TIC (\d+) sector (\d+)', identifier_string)
        if match:
            return int(match.group(1)), int(match.group(2))
        # Search for the human-readable TIC only version. E.g., "TIC 169480782"
        match = re.search(r'TIC (\d+)', identifier_string)
        if match:
            return int(match.group(1)), None
        # Search for the TESS obs_id version. E.g., "tess2018319095959-s0005-0000000278956474-0125-s"
        match = re.search(r'tess\d+-s(\d+)-(\d+)-\d+-s', identifier_string)
        if match:
            return int(match.group(2)), int(match.group(1))
        # Search for the TESS SPOC version. E.g., "hlsp_tess-spoc_tess_phot_0000000007583207-s0026_tess_v1_lc"
        match = re.search(r'hlsp_tess-spoc_tess_phot_(\d+)-s(\d+)_tess_v1', identifier_string)
        if match:
            return int(match.group(1)), int(match.group(2))
        # Raise an error if none of the patterns matched.
        raise ValueError(f'{identifier_string} does not match a known pattern to extract TIC ID and sector from.')


class TessTwoMinuteCadenceLightCurve(TessMissionLightCurve):
    pass


def ensure_native_byte_order(array: np.ndarray) -> np.ndarray:
    native_byte_order = '>' if sys.byteorder == 'big' else '<'
    if array.dtype.byteorder in ['|', '=', native_byte_order]:
        return array
    else:
        return array.byteswap().newbyteorder(native_byte_order)



"""
Code for vetting transit candidates.
"""
import pandas as pd

from ramjet.photometric_database.tess_target import TessTarget


class TransitVetter:
    """
    A class for vetting transit candidates.
    """
    radius_of_jupiter__solar_radii = 0.1028

    def is_transit_depth_for_target_physical_for_planet(self, target: TessTarget, transit_depth: float) -> bool:
        """
        Check if the depth of a transit is not too deep to be caused by a planet for a given target.

        :param target: The target whose parameters should be used.
        :param transit_depth: The transit depth in relative flux.
        :return: A boolean stating if the depth is physical for a planet.
                 False meaning the radius of transiting body is (likely) too large to be a planet.
        """
        transiting_body_radius = target.calculate_transiting_body_radius(transit_depth)
        planet_radius_threshold = 1.8 * self.radius_of_jupiter__solar_radii
        if transiting_body_radius < planet_radius_threshold:
            return True
        else:
            return False

    @staticmethod
    def has_no_nearby_likely_eclipsing_binary_background_targets(target: TessTarget) -> bool:
        """
        Checks if the target has likely nearby targets which may be eclipsing binaries showing up as the transit.

        :param target: The target of interest.
        :return: Whether or not there is at least one problematic nearby target.
        """
        nearby_threshold_arcseconds = 21  # 21 arcseconds is the size of the side of a TESS pixel.
        magnitude_difference_threshold = 5
        nearby_target_data_frame = target.retrieve_nearby_tic_targets()
        problematic_nearby_target_data_frame = nearby_target_data_frame.loc[
            (nearby_target_data_frame['TESS Mag'] < target.magnitude + magnitude_difference_threshold) &
            (nearby_target_data_frame['Separation (arcsec)'] < nearby_threshold_arcseconds)
        ]
        if problematic_nearby_target_data_frame.shape[0] == 0:
            return True
        else:
            return False

    @staticmethod
    def has_nearby_toi_targets(target: TessTarget) -> bool:
        """
        Checks if the target has nearby TOI targets which are likely to be where the transit is from.

        :param target: The target of interest.
        :return: Whether or not there is at least one problematic nearby target.
        """
        nearby_threshold_arcseconds = 31.5  # 1.5 TESS pixels.
        nearby_target_data_frame = target.retrieve_nearby_tic_targets()
        problematic_nearby_target_data_frame = nearby_target_data_frame.loc[
            (pd.notnull(nearby_target_data_frame['TOI'])) &
            (nearby_target_data_frame['Separation (arcsec)'] < nearby_threshold_arcseconds)
        ]
        if problematic_nearby_target_data_frame.shape[0] != 0:
            return True
        else:
            return False

    def get_maximum_physical_depth_for_planet_for_target(self, target: TessTarget,
                                                         allow_missing_contamination_ratio: bool = False) -> float:
        """
        Determines the maximum depth allowable for a given target for a transit to be caused by a planet.

        :param target: The target to check for.
        :param allow_missing_contamination_ratio: Allow for unknown contamination, which will then default to 0.
        :return: The maximum relative depth allowed.
        """
        maximum_planet_radius = 1.8 * self.radius_of_jupiter__solar_radii
        contamination_ratio = target.contamination_ratio
        if pd.isna(contamination_ratio):
            if allow_missing_contamination_ratio:
                contamination_ratio = 0
            else:
                raise ValueError(f'Contamination ratio {contamination_ratio} is not a number.')
        maximum_physical_depth = (maximum_planet_radius ** 2) / (
            (target.radius ** 2) * (1 + contamination_ratio))
        return maximum_physical_depth



"""
A model for the database entity of a TESS transiter.
"""
from peewee import AutoField, ForeignKeyField, FloatField, BooleanField

from ramjet.database.base_model import BaseModel
from ramjet.database.tess_target import TessTarget


class TessTransiter(BaseModel):
    """
    A database model for the database entity of a TESS transiter.
    """
    id = AutoField()
    target: TessTarget = ForeignKeyField(TessTarget)
    radius__solar_radii = FloatField(null=True)
    has_known_contamination_ratio = BooleanField(default=True)
    transit_epoch__btjd = FloatField()
    transit_period__days = FloatField()
    transit_duration__days = FloatField()
    transit_relative_depth = FloatField()



"""
A viewer for a CSV file containing a column of paths.
"""
from __future__ import annotations

import asyncio
import numpy as np
import pandas as pd
from typing import Union
from pathlib import Path
from bokeh.io import curdoc
from functools import partial
from bokeh.models import Button, Div, BoxAnnotation
from bokeh.document import Document
from bokeh.server.server import Server

from ramjet.analysis.transit_vetter import TransitVetter
from ramjet.analysis.viewer.light_curve_display import LightCurveDisplay
from ramjet.analysis.viewer.preloader import Preloader
from ramjet.analysis.viewer.view_entity import ViewEntity
from ramjet.photometric_database.tess_ffi_light_curve import TessFfiColumnName, TessFfiLightCurve


class Viewer:
    """
    A viewer for a CSV file containing a column of paths.
    """
    vetter = TransitVetter()

    def __init__(self):
        self.light_curve_display: Union[LightCurveDisplay, None] = None
        self.preloader: Union[Preloader, None] = None
        self.add_to_positives_button: Union[Button, None] = None
        self.previous_button: Union[Button, None] = None
        self.next_button: Union[Button, None] = None
        self.information_div: Union[Div, None] = None
        self.document: Union[Document, None] = None
        self.maximum_physical_depth_box: Union[BoxAnnotation, None] = None
        self.view_entity: Union[ViewEntity, None] = None

    async def update_view_entity_with_document_lock(self, view_entity: ViewEntity):
        """
        Updates the light curve display using the Bokeh document lock.

        :param view_entity: The view entity to update the display with.
        """
        light_curve = view_entity.light_curve
        self.document.add_next_tick_callback(partial(self.light_curve_display.update_from_light_curve,
                                                     light_curve=light_curve))
        self.document.add_next_tick_callback(partial(self.update_information_div_for_view_entity,
                                                     view_entity=view_entity))
        self.document.add_next_tick_callback(partial(self.add_physical_depth_range_annotation_to_light_curve_figure,
                                                     view_entity=view_entity))
        self.view_entity = view_entity

    async def add_physical_depth_range_annotation_to_light_curve_figure(self, view_entity: ViewEntity):
        unknown_radius = False
        maximum_depth = self.vetter.get_maximum_physical_depth_for_planet_for_target(
            view_entity.target, allow_missing_contamination_ratio=True)
        if np.isnan(maximum_depth):
            maximum_depth = 0.1
            unknown_radius = True
        self.maximum_physical_depth_box.bottom = 1 - maximum_depth
        if view_entity.has_exofop_dispositions:
            self.maximum_physical_depth_box.fill_color = 'red'
        elif unknown_radius:
            self.maximum_physical_depth_box.fill_color = 'yellow'
        else:
            self.maximum_physical_depth_box.fill_color = 'green'


    async def update_information_div_for_view_entity(self, view_entity: ViewEntity):
        self.information_div.text = (f'<h1 class="title">TIC {view_entity.light_curve.tic_id} ' +
                                     f'sector {view_entity.light_curve.sector}</h1>' +
                                     f'<p>Network confidence: {view_entity.confidence}</p>' +
                                     f'<p>Result index: {view_entity.index}</p>' +
                                     f'<p>Star radius (solar radii): {view_entity.target.radius}</p>')

    async def display_next_view_entity(self):
        """
        Moves to the next view entity.
        """
        next_view_entity = await self.preloader.increment()
        await self.update_view_entity_with_document_lock(next_view_entity)

    async def display_previous_view_entity(self):
        """
        Moves to the previous view entity.
        """
        previous_view_entity = await self.preloader.decrement()
        await self.update_view_entity_with_document_lock(previous_view_entity)

    def create_display_next_view_entity_task(self):
        """
        Creates the async task to move to the next light curve.
        """
        asyncio.create_task(self.display_next_view_entity())

    def create_display_previous_view_entity_task(self):
        """
        Creates the async task to move to the previous light curve.
        """
        asyncio.create_task(self.display_previous_view_entity())

    def create_light_curve_switching_buttons(self) -> (Button, Button):
        """
        Creates buttons for switching between light curves.
        """
        next_button = Button(label='Next target')
        next_button.on_click(self.create_display_next_view_entity_task)
        next_button.sizing_mode = 'stretch_width'
        previous_button = Button(label='Previous target')
        previous_button.on_click(self.create_display_previous_view_entity_task)
        previous_button.sizing_mode = 'stretch_width'
        return previous_button, next_button

    def create_add_to_positives_button(self) -> Button:
        add_to_positives_button = Button(label='Add to positives')
        add_to_positives_button.on_click(self.add_current_to_positives)
        add_to_positives_button.sizing_mode = 'stretch_width'
        return add_to_positives_button

    def add_current_to_positives(self):
        positives_csv_file_path = Path('positives.csv')
        positives_data_frame = pd.DataFrame({'tic_id': [self.view_entity.target.tic_id]})
        if positives_csv_file_path.exists():
            positives_data_frame.to_csv(positives_csv_file_path, mode='a', header=False, index=False)
        else:
            positives_data_frame.to_csv(positives_csv_file_path, index=False)

    @classmethod
    def from_csv_path(cls, bokeh_document: Document, csv_path: Path) -> Viewer:
        """
        Creates a viewer from a CSV path containing a light curve path column.

        :param bokeh_document: The Bokeh document to run the viewer in.
        :param csv_path: The path to the CSV file.
        :return: The viewer.
        """
        viewer = cls()
        viewer.document = bokeh_document
        viewer.csv_path = csv_path
        viewer.light_curve_display = LightCurveDisplay.for_columns(TessFfiColumnName.TIME__BTJD.value,
                                                                   TessFfiLightCurve().flux_column_names,
                                                                   flux_axis_label='Relative flux')
        viewer.light_curve_display.exclude_outliers_from_zoom = True
        viewer.maximum_physical_depth_box = BoxAnnotation(bottom=1-0.01, top=1, fill_alpha=0.1, fill_color='green')
        viewer.light_curve_display.figure.add_layout(viewer.maximum_physical_depth_box)
        viewer.add_to_positives_button = viewer.create_add_to_positives_button()
        bokeh_document.add_root(viewer.add_to_positives_button)
        viewer.previous_button, viewer.next_button = viewer.create_light_curve_switching_buttons()
        bokeh_document.add_root(viewer.previous_button)
        bokeh_document.add_root(viewer.next_button)
        viewer.information_div = Div()
        viewer.information_div.sizing_mode = 'stretch_width'
        bokeh_document.add_root(viewer.information_div)
        bokeh_document.add_root(viewer.light_curve_display.figure)
        loop = asyncio.get_running_loop()
        loop.create_task(viewer.start_preloader(csv_path))
        return viewer

    async def start_preloader(self, csv_path):
        """
        Starts the light curve preloader.
        """
        self.preloader = await Preloader.from_csv_path(csv_path, starting_index=0)
        initial_view_entity = self.preloader.current_view_entity
        await self.update_view_entity_with_document_lock(initial_view_entity)


def application(bokeh_document: Document):
    """
    The application to run from the Tornado server.

    :param bokeh_document: The Bokeh document to run the viewer in.
    """
    csv_path = Path('/Users/golmsche/Desktop/infer results 2020-10-09-13-21-21.csv')
    Viewer.from_csv_path(bokeh_document, csv_path)


if __name__ == '__main__':
    document = curdoc()
    server = Server({'/': application}, port=5010)
    server.start()
    server.io_loop.add_callback(server.show, "/")
    server.io_loop.start()



import copy
import math
import re
import shutil
import socket
from enum import Enum
from functools import partial
from pathlib import Path
from typing import List, Iterable, Self, Tuple, TypeVar, Iterator, Callable, Any

import numpy as np
import numpy.typing as npt
import torch
from filelock import FileLock
from scipy import stats
from scipy.interpolate import interp1d
from torch import Tensor
from torch.utils.data import IterableDataset

from qusi.light_curve import LightCurve, remove_nan_flux_data_points_from_light_curve
from qusi.light_curve_collection import LabeledLightCurveCollection
from qusi.light_curve_observation import LightCurveObservation, remove_nan_flux_data_points_from_light_curve_observation
from qusi.light_curve_transforms import from_light_curve_observation_to_fluxes_array_and_label_array, \
    pair_array_to_tensor


class LightCurveDataset(IterableDataset):
    """
    A dataset of light curve data.
    """

    def __init__(self,
                 standard_light_curve_collections: List[LabeledLightCurveCollection],
                 injectee_light_curve_collections: List[LabeledLightCurveCollection],
                 injectable_light_curve_collections: List[LabeledLightCurveCollection],
                 post_injection_transform: Callable[[Any], Any],
                 ):
        self.standard_light_curve_collections: List[LabeledLightCurveCollection] = standard_light_curve_collections
        self.injectee_light_curve_collections: List[LabeledLightCurveCollection] = injectee_light_curve_collections
        self.injectable_light_curve_collections: List[LabeledLightCurveCollection] = injectable_light_curve_collections
        if len(self.standard_light_curve_collections) == 0 and len(self.injectee_light_curve_collections) == 0:
            raise ValueError('Either the standard or injectee light curve collection lists must not be empty. '
                             'Both were empty.')
        self.post_injection_transform: Callable[[Any], Any] = post_injection_transform

    def __iter__(self):
        base_light_curve_collection_iter_and_type_pairs: \
            List[Tuple[Iterator[LightCurveObservation], LightCurveCollectionType]] = []
        injectee_collections = copy.copy(self.injectee_light_curve_collections)
        for standard_collection in self.standard_light_curve_collections:
            if standard_collection in injectee_collections:
                base_light_curve_collection_iter_and_type_pairs.append(
                    (loop_iter_function(standard_collection.observation_iter),
                     LightCurveCollectionType.STANDARD_AND_INJECTEE))
                injectee_collections.remove(standard_collection)
            else:
                base_light_curve_collection_iter_and_type_pairs.append(
                    (loop_iter_function(standard_collection.observation_iter), LightCurveCollectionType.STANDARD))
        for injectee_collection in injectee_collections:
            base_light_curve_collection_iter_and_type_pairs.append(
                (loop_iter_function(injectee_collection.observation_iter), LightCurveCollectionType.INJECTEE))
        injectable_light_curve_collection_iters: List[Iterator[LightCurveObservation]] = []
        for injectable_collection in self.injectable_light_curve_collections:
            injectable_light_curve_collection_iters.append(loop_iter_function(injectable_collection.observation_iter))
        while True:
            for base_light_curve_collection_iter_and_type_pair in base_light_curve_collection_iter_and_type_pairs:
                base_collection_iter, collection_type = base_light_curve_collection_iter_and_type_pair
                if collection_type in [LightCurveCollectionType.STANDARD,
                                       LightCurveCollectionType.STANDARD_AND_INJECTEE]:
                    # TODO: Preprocessing step should be here. Or maybe that should all be on the light curve collection
                    #  as well? Or passed in somewhere else?
                    standard_light_curve = next(base_collection_iter)
                    transformed_standard_light_curve = self.post_injection_transform(standard_light_curve)
                    yield transformed_standard_light_curve
                if collection_type in [LightCurveCollectionType.INJECTEE,
                                       LightCurveCollectionType.STANDARD_AND_INJECTEE]:
                    for injectable_light_curve_collection_iter in injectable_light_curve_collection_iters:
                        injectable_light_curve = next(injectable_light_curve_collection_iter)
                        injectee_light_curve = next(base_collection_iter)
                        injected_light_curve = inject_light_curve(injectee_light_curve, injectable_light_curve)
                        transformed_injected_light_curve = self.post_injection_transform(injected_light_curve)
                        yield transformed_injected_light_curve

    @classmethod
    def new(cls,
            standard_light_curve_collections: List[LabeledLightCurveCollection] | None = None,
            injectee_light_curve_collections: List[LabeledLightCurveCollection] | None = None,
            injectable_light_curve_collections: List[LabeledLightCurveCollection] | None = None,
            post_injection_transform: Callable[[Any], Any] | None = None,
            ) -> Self:
        if standard_light_curve_collections is None and injectee_light_curve_collections is None:
            raise ValueError('Either the standard or injectee light curve collection lists must be specified. '
                             'Both were `None`.')
        if standard_light_curve_collections is None:
            standard_light_curve_collections = []
        if injectee_light_curve_collections is None:
            injectee_light_curve_collections = []
        if injectable_light_curve_collections is None:
            injectable_light_curve_collections = []
        if post_injection_transform is None:
            post_injection_transform = partial(default_light_curve_observation_post_injection_transform, length=2500)
        instance = cls(standard_light_curve_collections=standard_light_curve_collections,
                       injectee_light_curve_collections=injectee_light_curve_collections,
                       injectable_light_curve_collections=injectable_light_curve_collections,
                       post_injection_transform=post_injection_transform)
        return instance


def inject_light_curve(injectee_observation: LightCurveObservation, injectable_observation: LightCurveObservation
                       ) -> LightCurveObservation:
    fluxes_with_injected_signal, _, _ = inject_signal_into_light_curve_with_intermediates(
        light_curve_times=injectee_observation.light_curve.times,
        light_curve_fluxes=injectee_observation.light_curve.fluxes,
        signal_times=injectable_observation.light_curve.times,
        signal_magnifications=injectable_observation.light_curve.fluxes,
        out_of_bounds_injection_handling_method=OutOfBoundsInjectionHandlingMethod.RANDOM_INJECTION_LOCATION,
        baseline_flux_estimation_method=BaselineFluxEstimationMethod.MEDIAN
    )
    injected_light_curve = LightCurve.new(times=injectee_observation.light_curve.times,
                                          fluxes=injectee_observation.light_curve.fluxes)
    injected_observation = LightCurveObservation.new(light_curve=injected_light_curve,
                                                     label=injectable_observation.label)
    return injected_observation


def is_injected_dataset(dataset: LightCurveDataset):
    return len(dataset.injectee_light_curve_collections) > 0


def contains_injected_dataset(datasets: List[LightCurveDataset]):
    for dataset in datasets:
        if is_injected_dataset(dataset):
            return True
    return False


def interleave_infinite_iterators(*infinite_iterators: Iterator):
    while True:
        for iterator in infinite_iterators:
            yield next(iterator)


T = TypeVar('T')


def loop_iter_function(iter_function: Callable[[], Iterable[T]]) -> Iterator[T]:
    while True:
        iterator = iter_function()
        for element in iterator:
            yield element


class ObservationType(Enum):
    STANDARD = 'standard'
    INJECTEE = 'injectee'


class LightCurveCollectionType(Enum):
    STANDARD = 'standard'
    INJECTEE = 'injectee'
    STANDARD_AND_INJECTEE = 'standard_and_injectee'


class InterleavedDataset(IterableDataset):
    def __init__(self, *datasets: IterableDataset):
        self.datasets: Tuple[IterableDataset] = datasets

    @classmethod
    def new(cls, *datasets: IterableDataset):
        instance = cls(*datasets)
        return instance

    def __iter__(self):
        # noinspection PyTypeChecker
        dataset_iterators = list(map(iter, self.datasets))
        return interleave_infinite_iterators(*dataset_iterators)


class ConcatenatedIterableDataset(IterableDataset):
    def __init__(self, *datasets: IterableDataset):
        self.datasets: Tuple[IterableDataset] = datasets

    @classmethod
    def new(cls, *datasets: IterableDataset):
        instance = cls(*datasets)
        return instance

    def __iter__(self):
        for dataset in self.datasets:
            for element in dataset:
                yield element


class LimitedIterableDataset(IterableDataset):
    def __init__(self, dataset: IterableDataset, limit: int):
        self.dataset: IterableDataset = dataset
        self.limit: int = limit

    @classmethod
    def new(cls, dataset: IterableDataset, limit: int):
        instance = cls(dataset, limit)
        return instance

    def __iter__(self):
        count = 0
        for element in self.dataset:
            yield element
            count += 1
            if count >= self.limit:
                break


def default_light_curve_observation_post_injection_transform(x: LightCurveObservation, length: int) -> (Tensor, Tensor):
    x = remove_nan_flux_data_points_from_light_curve_observation(x)
    x = from_light_curve_observation_to_fluxes_array_and_label_array(x)
    x = make_fluxes_and_label_array_uniform_length(x, length=length)
    x = pair_array_to_tensor(x)
    x = (normalize_tensor_by_modified_z_score(x[0]), x[1])
    return x

def default_light_curve_post_injection_transform(x: LightCurve, length: int) -> (Tensor, Tensor):
    x = remove_nan_flux_data_points_from_light_curve(x)
    x = x.fluxes
    x = make_uniform_length(x, length=length)
    x = torch.tensor(x)
    x = normalize_tensor_by_modified_z_score(x)
    return x

def normalize_tensor_by_modified_z_score(tensor: Tensor) -> Tensor:
    median = torch.median(tensor)
    deviation_from_median = tensor - median
    absolute_deviation_from_median = torch.abs(deviation_from_median)
    median_absolute_deviation_from_median = torch.median(absolute_deviation_from_median)
    modified_z_score = 0.6745 * deviation_from_median / median_absolute_deviation_from_median
    return modified_z_score


def make_fluxes_and_label_array_uniform_length(arrays: Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]],
                                               length: int, randomize: bool = True) -> (np.ndarray, np.ndarray):
    fluxes, label = arrays
    uniform_length_times = make_uniform_length(fluxes, length=length, randomize=randomize)
    return uniform_length_times, label


def make_uniform_length(example: np.ndarray, length: int, randomize: bool = True) -> np.ndarray:
    """Makes the example a specific length, by clipping those too large and repeating those too small."""
    assert len(example.shape) in [1, 2]  # Only tested for 1D and 2D cases.
    if randomize:
        example = randomly_roll_elements(example)
    if example.shape[0] == length:
        pass
    elif example.shape[0] > length:
        example = example[:length]
    else:
        elements_to_repeat = length - example.shape[0]
        if len(example.shape) == 1:
            example = np.pad(example, (0, elements_to_repeat), mode='wrap')
        else:
            example = np.pad(example, ((0, elements_to_repeat), (0, 0)), mode='wrap')
    return example


def randomly_roll_elements(example: np.ndarray) -> np.ndarray:
    """Randomly rolls the elements."""
    example = np.roll(example, np.random.randint(example.shape[0]), axis=0)
    return example


class OutOfBoundsInjectionHandlingMethod(Enum):
    """
    An enum of approaches for handling cases where the injectable signal is shorter than the injectee signal.
    """
    ERROR = 'error'
    REPEAT_SIGNAL = 'repeat_signal'
    RANDOM_INJECTION_LOCATION = 'random_inject_location'


class BaselineFluxEstimationMethod(Enum):
    """
    An enum of to designate the type of baseline flux estimation method to use during training.
    """
    MEDIAN = 'median'
    MEDIAN_ABSOLUTE_DEVIATION = 'median_absolute_deviation'


def inject_signal_into_light_curve_with_intermediates(
        light_curve_times: npt.NDArray[np.float64],
        light_curve_fluxes: npt.NDArray[np.float64],
        signal_times: npt.NDArray[np.float64],
        signal_magnifications: npt.NDArray[np.float64],
        out_of_bounds_injection_handling_method: OutOfBoundsInjectionHandlingMethod =
        OutOfBoundsInjectionHandlingMethod.ERROR,
        baseline_flux_estimation_method: BaselineFluxEstimationMethod = BaselineFluxEstimationMethod.MEDIAN
) -> (npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]):
    """
    Injects a synthetic magnification signal into real light curve fluxes.

    :param light_curve_times: The times of the flux observations of the light curve.
    :param light_curve_fluxes: The fluxes of the light curve to be injected into.
    :param signal_times: The times of the synthetic magnifications.
    :param signal_magnifications: The synthetic magnifications to inject.
    :param out_of_bounds_injection_handling_method: The method to use to handle out of bounds injection.
    :param baseline_flux_estimation_method: The method to use to estimate the baseline flux of the light curve
                                            for scaling the signal magnifications.
    :return: The fluxes with the injected signal, the offset signal times, and the signal flux.
    """
    minimum_light_curve_time = np.min(light_curve_times)
    relative_light_curve_times = light_curve_times - minimum_light_curve_time
    relative_signal_times = signal_times - np.min(signal_times)
    signal_time_length = np.max(relative_signal_times)
    light_curve_time_length = np.max(relative_light_curve_times)
    time_length_difference = light_curve_time_length - signal_time_length
    signal_start_offset = (np.random.random() * time_length_difference) + minimum_light_curve_time
    offset_signal_times = relative_signal_times + signal_start_offset
    if baseline_flux_estimation_method == BaselineFluxEstimationMethod.MEDIAN_ABSOLUTE_DEVIATION:
        baseline_flux = stats.median_abs_deviation(light_curve_fluxes)
        baseline_to_median_absolute_deviation_ratio = 10  # Arbitrarily chosen to give a reasonable scale.
        baseline_flux *= baseline_to_median_absolute_deviation_ratio
    else:
        baseline_flux = np.median(light_curve_fluxes)
    signal_fluxes = (signal_magnifications * baseline_flux) - baseline_flux
    if out_of_bounds_injection_handling_method is OutOfBoundsInjectionHandlingMethod.RANDOM_INJECTION_LOCATION:
        signal_flux_interpolator = interp1d(offset_signal_times, signal_fluxes, bounds_error=False, fill_value=0)
    elif (out_of_bounds_injection_handling_method is OutOfBoundsInjectionHandlingMethod.REPEAT_SIGNAL and
          time_length_difference > 0):
        before_signal_gap = signal_start_offset - minimum_light_curve_time
        after_signal_gap = time_length_difference - before_signal_gap
        minimum_signal_time_step = np.min(np.diff(offset_signal_times))
        before_repeats_needed = math.ceil(before_signal_gap / (signal_time_length + minimum_signal_time_step))
        after_repeats_needed = math.ceil(after_signal_gap / (signal_time_length + minimum_signal_time_step))
        repeated_signal_fluxes = np.tile(signal_fluxes, before_repeats_needed + 1 + after_repeats_needed)
        repeated_signal_times = None
        for repeat_index in range(-before_repeats_needed, after_repeats_needed + 1):
            repeat_signal_start_offset = (signal_time_length + minimum_signal_time_step) * repeat_index
            if repeated_signal_times is None:
                repeated_signal_times = offset_signal_times + repeat_signal_start_offset
            else:
                repeat_index_signal_times = offset_signal_times + repeat_signal_start_offset
                repeated_signal_times = np.concatenate([repeated_signal_times, repeat_index_signal_times])
        signal_flux_interpolator = interp1d(repeated_signal_times, repeated_signal_fluxes, bounds_error=True)
    else:
        signal_flux_interpolator = interp1d(offset_signal_times, signal_fluxes, bounds_error=True)
    interpolated_signal_fluxes = signal_flux_interpolator(light_curve_times)
    fluxes_with_injected_signal = light_curve_fluxes + interpolated_signal_fluxes
    return fluxes_with_injected_signal, offset_signal_times, signal_fluxes


def move_path_to_nvme(path: Path) -> Path:
    match = re.match(r"gpu\d{3}", socket.gethostname())
    if match is not None:
        nvme_path = Path("/lscratch/golmsche").joinpath(path)
        if not nvme_path.exists():
            nvme_path.parent.mkdir(exist_ok=True, parents=True)
            nvme_lock_path = nvme_path.parent.joinpath(nvme_path.name + '.lock')
            lock = FileLock(str(nvme_lock_path))
            with lock.acquire():
                if not nvme_path.exists():
                    nvme_tmp_path = nvme_path.parent.joinpath(nvme_path.name + '.tmp')
                    shutil.copy(path, nvme_tmp_path)
                    nvme_tmp_path.rename(nvme_path)
        return nvme_path
    else:
        return path


"""Code for a base generalized database for photometric data to be subclassed."""
import math
import shutil
from abc import ABC
from pathlib import Path
from typing import List, Union, Callable, Iterable, Tuple

import numpy as np
import numpy.typing as npt



def preprocess_times(light_curve_array: np.ndarray) -> None:
    """
    Preprocesses the times of the light curve.

    :param light_curve_array: The light curve array to preprocess.
    :return: The light curve array with the times preprocessed.
    """
    times = light_curve_array[:, 0]
    light_curve_array[:, 0] = calculate_time_differences(times)


def make_times_and_fluxes_array_uniform_length(arrays: Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]], length: int, randomize: bool = True) -> (np.ndarray, np.ndarray):
    times, fluxes = arrays
    light_curve_array = np.stack([times, fluxes], axis=-1)
    uniform_length_light_curve_array = make_uniform_length(light_curve_array, length=length, randomize=randomize)
    return uniform_length_light_curve_array[:, 0], uniform_length_light_curve_array[:, 1]


def make_fluxes_and_label_array_uniform_length(arrays: Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]], length: int, randomize: bool = True) -> (np.ndarray, np.ndarray):
    times, label = arrays
    uniform_length_times = make_uniform_length(times, length=length, randomize=randomize)
    return uniform_length_times, label


def make_uniform_length(example: np.ndarray, length: int, randomize: bool = True) -> np.ndarray:
    """Makes the example a specific length, by clipping those too large and repeating those too small."""
    assert len(example.shape) in [1, 2]  # Only tested for 1D and 2D cases.
    if randomize:
        example = randomly_roll_elements(example)
    if example.shape[0] == length:
        pass
    elif example.shape[0] > length:
        example = example[:length]
    else:
        elements_to_repeat = length - example.shape[0]
        if len(example.shape) == 1:
            example = np.pad(example, (0, elements_to_repeat), mode='wrap')
        else:
            example = np.pad(example, ((0, elements_to_repeat), (0, 0)), mode='wrap')
    return example


class LightCurveDatabase(ABC):
    """A base generalized database for photometric data to be subclassed."""

    def __init__(self, data_directory='data'):
        self.time_steps_per_example = 16000
        self.data_directory: Path = Path(data_directory)
        self.validation_ratio: float = 0.2
        self.batch_size: int = 100
        self.time_steps_per_example: int
        self.number_of_parallel_processes_per_map: int = 16
        self.include_time_as_channel: bool = False
        self.include_flux_errors_as_channel: bool = False

    @property
    def window_shift(self) -> int:
        """
        How much the window shifts for a windowed batch set.

        :return: The window shift size.
        """
        return self.batch_size // 10

    def clear_data_directory(self):
        """
        Empties the data directory.
        """
        if self.data_directory.exists():
            shutil.rmtree(self.data_directory)
        self.create_data_directories()

    def create_data_directories(self):
        """
        Creates the data directories to be used by the database.
        """
        self.data_directory.mkdir(parents=True, exist_ok=True)

    def normalize_fluxes(self, light_curve: np.ndarray) -> None:
        """
        Normalizes the flux channel of the light curve in-place.

        :param light_curve: The light curve whose flux channel should be normalized.
        :return: The light curve with the flux channel normalized.
        """
        if self.include_time_as_channel:
            if self.include_flux_errors_as_channel:
                assert light_curve.shape[1] == 3
                light_curve[:, 1], light_curve[:, 2] = normalize_on_percentiles_with_errors(
                    light_curve[:, 1], light_curve[:, 2])
            else:
                assert light_curve.shape[1] == 2
                light_curve[:, 1] = normalize_on_percentiles(light_curve[:, 1])
        else:
            assert light_curve.shape[1] == 1
            light_curve[:, 0] = normalize_on_percentiles(light_curve[:, 0])

    def build_light_curve_array(self, fluxes: np.ndarray, times: Union[np.ndarray, None] = None,
                                flux_errors: Union[np.ndarray, None] = None):
        """
        Builds the light curve array based on the components required for the specific database setup.

        :param fluxes: The fluxes of the light curve.
        :param times: The optional times of the light curve.
        :param flux_errors: The optional flux errors of the light curve.
        :return: The constructed light curve array.
        """
        if self.include_flux_errors_as_channel:
            if not self.include_time_as_channel:
                raise NotImplementedError
            light_curve = np.stack([times, fluxes, flux_errors], axis=-1)
        elif self.include_time_as_channel:
            light_curve = np.stack([times, fluxes], axis=-1)
        else:
            light_curve = np.expand_dims(fluxes, axis=-1)
        return light_curve

    def preprocess_light_curve(self, light_curve: np.ndarray, evaluation_mode: bool = False) -> np.ndarray:
        """
        Preprocessing for the light curve.

        :param light_curve: The light curve array to preprocess.
        :param evaluation_mode: If the preprocessing should be consistent for evaluation.
        :return: The preprocessed flux array.
        """
        if not evaluation_mode:
            light_curve = remove_random_elements(light_curve)
        light_curve = make_uniform_length(light_curve, self.time_steps_per_example,
                                          randomize=not evaluation_mode)
        self.normalize_fluxes(light_curve)
        if self.include_time_as_channel:
            preprocess_times(light_curve)
        return light_curve


def normalize_log_0_to_1(light_curve: np.ndarray) -> np.ndarray:
    """Normalizes from 0 to 1 on the logarithm of the light curve."""
    light_curve -= np.min(light_curve)
    light_curve = np.log1p(light_curve)
    array_max = np.max(light_curve)
    if array_max != 0:
        light_curve /= array_max
    return light_curve


def normalize_on_percentiles(array: np.ndarray) -> np.ndarray:
    """
    Normalizes an array using percentiles. The 10th percentile is normalized to -1, the 90th to 1.

    :param array: The array to be normalized.
    :return: The normalized array.
    """
    percentile_10 = np.percentile(array, 10)
    percentile_90 = np.percentile(array, 90)
    percentile_difference = percentile_90 - percentile_10
    if percentile_difference == 0:
        normalized_array = np.zeros_like(array)
    else:
        normalized_array = ((array - percentile_10) / (percentile_difference / 2)) - 1
    return normalized_array


def normalize_on_percentiles_with_errors(array: np.ndarray, array_errors: np.ndarray) -> (np.ndarray, np.ndarray):
    """
    Normalizes an array using percentiles. The 10th percentile is normalized to -1, the 90th to 1.
    Scales the errors by the corresponding scaling factor.
    """
    percentile_10 = np.percentile(array, 10)
    percentile_90 = np.percentile(array, 90)
    percentile_difference = percentile_90 - percentile_10
    if percentile_difference == 0:
        normalized_array = np.zeros_like(array)
        normalized_array_errors = np.zeros_like(array_errors)
    else:
        normalized_array = ((array - percentile_10) / (percentile_difference / 2)) - 1
        normalized_array_errors = array_errors / (percentile_difference / 2)
    return normalized_array, normalized_array_errors


def shuffle_in_unison(a, b, seed=None):
    """Shuffle two arrays in unison."""
    if seed is not None:
        np.random.seed(seed)
    indexes = np.random.permutation(len(a))
    return np.array(a)[indexes], np.array(b)[indexes]


def remove_random_elements(light_curve: np.ndarray, ratio: float = 0.01) -> np.ndarray:
    """Removes random values from the light_curve."""
    light_curve_length = light_curve.shape[0]
    max_values_to_remove = int(light_curve_length * ratio)
    if max_values_to_remove != 0:
        values_to_remove = np.random.randint(max_values_to_remove)
    else:
        values_to_remove = 0
    random_indexes = np.random.choice(range(light_curve_length), values_to_remove, replace=False)
    return np.delete(light_curve, random_indexes, axis=0)


def randomly_roll_elements(example: np.ndarray) -> np.ndarray:
    """Randomly rolls the elements."""
    example = np.roll(example, np.random.randint(example.shape[0]), axis=0)
    return example


def extract_shuffled_chunk_and_remainder(array_to_extract_from: Union[List, np.ndarray], chunk_ratio: float,
                                         chunk_to_extract_index: int = 0) -> (np.ndarray, np.ndarray):
    """
    Shuffles an array, extracts a chunk of the data, and returns the chunk and remainder of the array.

    :param array_to_extract_from: The array to process.
    :param chunk_ratio: The number of equal size chunks to split the array into before extracting one.
    :param chunk_to_extract_index: The index of the chunk to extract out of all chunks.
    :return: The chunk which is extracted, and the remainder of the array excluding the chunk.
    """
    np.random.seed(0)
    np.random.shuffle(array_to_extract_from)
    number_of_chunks = int(1 / chunk_ratio)
    chunks = np.array_split(array_to_extract_from, number_of_chunks)
    extracted_chunk = chunks[chunk_to_extract_index]
    remaining_chunks = np.delete(chunks, chunk_to_extract_index, axis=0)
    remainder = np.concatenate(remaining_chunks)
    return extracted_chunk, remainder


def calculate_time_differences(times: np.ndarray) -> np.ndarray:
    """
    Calculates the differences between an array of time, doubling up the first element to make the length the same.

    :param times: The times to difference.
    :return: The time differences.
    """
    difference_times = np.diff(times)
    difference_times = np.insert(difference_times, 0, difference_times[0], axis=0)
    return difference_times



"""
Code for converting a column name to a display name.
"""

import re


def convert_column_name_to_display_name(column_name: str) -> str:
    """
    Add method to convert from column style names to human-readable display names. Column style names should be
    snake_case with a double underscore before the units (if units are relevant). The display name will have the
    first letter capitalized, spaces between words, and the units in parentheses at the end.
    Example: `transit_duration__days` -> `Transit duration (days)`

    :param column_name: The column whose name should be converted.
    :return: The display version of the name.
    """
    display_name = re.sub(r'__(.*)', r' (\g<1>)', column_name)  # Move units into parentheses.
    display_name = display_name.replace('_', ' ')
    specific_replacements = {
        r'tic id': 'TIC ID',
        r'\(btjd\)': '(BTJD)',
        r'\(bjd\)': '(BJD)',
        r'\(jd\)': '(JD)',
        r'pdcsap': 'PDCSAP',
        r'(^|\s)sap($|\s)': r'\g<1>SAP\g<2>',  # SAP avoiding possible cases where `sap` is part of a real word.
    }
    for pattern, replacement in specific_replacements.items():
        display_name = re.sub(pattern, replacement, display_name)
    display_name = display_name[0].upper() + display_name[1:]  # Make first letter uppercase.
    return display_name



import sys
from pathlib import Path

from bokeh import plotting
from bokeh.io import curdoc

from ramjet.analysis.light_curve_folding_vizualizer.viewer import Viewer
from ramjet.photometric_database.tess_ffi_light_curve import TessFfiLightCurve


def main():
    plotting.output_notebook.__doc__ = ''
    document = curdoc()
    light_curve_path = Path(sys.argv[1])
    light_curve = TessFfiLightCurve.from_path(light_curve_path)
    Viewer(document, light_curve, title=light_curve_path.name)


if __name__ == '__main__' or 'bokeh_app_' in __name__:
    main()



import torch
from torch import Tensor
from torch.nn import Module, Linear, Sigmoid


class SingleDenseLayerBinaryClassificationModel(Module):
    def __init__(self, input_size: int):
        super().__init__()
        self.dense = Linear(in_features=input_size, out_features=1)
        self.activation = Sigmoid()

    def forward(self, x: Tensor):
        x = self.dense(x)
        x = self.activation(x)
        x = torch.reshape(x, (-1,))
        return x



"""
Code for interacting with MOA light curve files and metadata.
"""
import numpy as np
import pandas as pd
from collections import defaultdict
from typing import List, Dict, Union
from pathlib import Path

import requests
from bs4 import BeautifulSoup


class MoaDataInterface:
    """
    A class for interacting with MOA light curve files and metadata.
    """
    survey_tag_to_path_list_dictionary_: Union[Dict[str, List[Path]], None] = None
    no_tag_string = 'no_tag'
    all_survey_tags = ['c', 'cf', 'cp', 'cw', 'cs', 'cb', 'v', 'n', 'nr', 'm', 'j', no_tag_string]

    @property
    def survey_tag_to_path_list_dictionary(self) -> Dict[str, List[Path]]:
        """
        Property allowing the survey tag to path list dictionary to only be loaded once.

        :return: The survey tag to path list dictionary
        """
        if self.survey_tag_to_path_list_dictionary_ is None:
            takahiro_sumi_nine_year_events_data_frame = self.read_corrected_nine_year_events_table_as_data_frame(
                Path('data/moa_microlensing/moa9yr_events_oct2018.txt'))
            self.survey_tag_to_path_list_dictionary_ = self.group_paths_by_tag_in_events_data_frame(
                list(Path('data/moa_microlensing').glob('**/*.cor.feather')), takahiro_sumi_nine_year_events_data_frame)
        return self.survey_tag_to_path_list_dictionary_

    @staticmethod
    def read_corrected_nine_year_events_table_as_data_frame(path: Path) -> pd.DataFrame:
        """
        Reads Takahiro Sumi's 9-year events table as a Pandas data frame, correcting for updates that appear on Yuki
        Hirao's website of the events (http://iral2.ess.sci.osaka-u.ac.jp/~moa/anomaly/9year/).

        :param path: The path to the events table file.
        :return: The data frame.
        """
        data_frame = MoaDataInterface.read_takahiro_sumi_nine_year_events_table_as_data_frame(path)
        yuki_hirao_data_frame = MoaDataInterface.get_yuki_hirao_events_data_frame()
        yuki_hirao_tag_data_frame = yuki_hirao_data_frame.filter(['tag'])
        data_frame.update(yuki_hirao_tag_data_frame)
        yuki_hirao_class_data_frame = yuki_hirao_data_frame.filter(['class'])
        data_frame: pd.DataFrame = data_frame.merge(yuki_hirao_class_data_frame, how='left', left_index=True,
                                                    right_index=True)
        data_frame.loc[(data_frame['class'].str.contains('b') |
                        data_frame['class'].str.contains('d') |
                        data_frame['class'].str.contains('p')) &
                       (data_frame['tag'] != 'cb'), 'tag'] = 'cb'
        data_frame.loc[(data_frame['tag'] == 'cb') &
                       (data_frame['class'].isin(['s'])), 'tag'] = 'c'  # Yuki Hirao labeled as single.
        data_frame.loc[(data_frame['tag'] == 'cb') &
                       (data_frame['class'].isin(['c'])), 'tag'] = 'n'  # Yuki Hirao labeled as cataclysmic variable.
        data_frame = data_frame.drop(columns='class')
        return data_frame

    @staticmethod
    def read_takahiro_sumi_nine_year_events_table_as_data_frame(path: Path) -> pd.DataFrame:
        """
        Reads Takahiro Sumi's 9-year events table as a Pandas data frame.

        :param path: The path to the events table file.
        :return: The data frame.
        """
        named_column_names = ['field', 'clr', 'chip', 'subfield', 'id', 'tag', 'x', 'y']
        # The number of columns in the file are inconsistent, so here we add extra unnamed columns to match the
        # largest number of columns in any row.
        largest_column_count = 33
        unnamed_column_names = [f'unnamed{index}' for index in range(largest_column_count - len(named_column_names))]
        column_names = named_column_names + unnamed_column_names
        data_frame = pd.read_csv(path, comment='#', names=column_names, delim_whitespace=True, skipinitialspace=True,
                                 skiprows=23)
        data_frame = data_frame.set_index(['field', 'clr', 'chip', 'subfield', 'id'], drop=False)
        data_frame = data_frame.sort_index()
        return data_frame

    @staticmethod
    def get_yuki_hirao_events_data_frame() -> pd.DataFrame:
        """
        Loads the events data from Yuki Hirao's website of events
        (http://iral2.ess.sci.osaka-u.ac.jp/~moa/anomaly/9year/).

        :return: The data frame of the events.
        """
        url = 'http://iral2.ess.sci.osaka-u.ac.jp/~moa/anomaly/9year/'
        page = requests.get(url)
        soup = BeautifulSoup(page.content, 'lxml')
        tbl = soup.find("table")
        events_data_frame = pd.read_html(str(tbl))[0]
        events_data_frame[['field', 'clr', 'chip', 'subfield', 'id']] = events_data_frame['MOA INTERNAL ID'].str.split(
            '-', 4, expand=True)
        events_data_frame['chip'] = events_data_frame['chip'].astype(int)
        events_data_frame['subfield'] = events_data_frame['subfield'].astype(int)
        events_data_frame['id'] = events_data_frame['id'].astype(int)
        events_data_frame = events_data_frame.set_index(['field', 'clr', 'chip', 'subfield', 'id'], drop=False)
        events_data_frame = events_data_frame.sort_index()
        return events_data_frame

    def get_tag_for_path_from_data_frame(self, path: Path, events_data_frame: pd.DataFrame) -> str:
        """
        Gets the event tag of a light curve from the events data frame.

        :param path: The path of the light curve whose event tag should be retrieved.
        :param events_data_frame: Takahiro Sumi's 9-year events data frame.
        :return: The string of the tag of the event. None if no tag exists.
        """
        file_name = path.name
        file_name_without_extension = file_name.split('.')[0]
        moa_identifier = file_name_without_extension.split('_')[-1]  # Remove duplicate identifier string.
        field, clr, chip_string, subfield_string, id_string = moa_identifier.split('-')
        chip, subfield, id_ = int(chip_string), int(subfield_string), int(id_string)
        try:
            row = events_data_frame.loc[(field, clr, chip, subfield, id_)]
            tag = row['tag']
            return tag
        except KeyError:
            return self.no_tag_string

    def group_paths_by_tag_in_events_data_frame(self, paths: List[Path], events_data_frame: pd.DataFrame
                                                ) -> Dict[str, List[Path]]:
        """
        Groups paths into a dictionary based on their tags.

        :param paths: The paths to group.
        :param events_data_frame: The events data frame to look for a tag in.
        :return:
        """
        tag_path_list_dictionary = defaultdict(list)
        for path in paths:
            tag = self.get_tag_for_path_from_data_frame(path, events_data_frame)
            tag_path_list_dictionary[tag].append(path)
        return tag_path_list_dictionary



from typing import Union, List

from ramjet.data_interface.moa_data_interface import MoaDataInterface
from ramjet.photometric_database.derived.moa_survey_light_curve_collection import MoaSurveyLightCurveCollection
from ramjet.photometric_database.standard_and_injected_light_curve_database import StandardAndInjectedLightCurveDatabase


class MoaSurveyBalancedTagDatabase(StandardAndInjectedLightCurveDatabase):
    """
    A database to train a network to find MOA events, balancing the training of each tag.
    """
    moa_data_interface = MoaDataInterface()

    def __init__(self):
        super().__init__()
        self.number_of_label_values = 2
        self.number_of_parallel_processes_per_map = 3
        self.time_steps_per_example = 18000
        self.training_standard_light_curve_collections = self.create_collection_for_each_tag(
            dataset_splits=list(range(8)))
        self.validation_standard_light_curve_collections = self.create_collection_for_each_tag(dataset_splits=[8])
        self.inference_light_curve_collections = self.create_collection_for_each_tag(dataset_splits=[9])

    def create_collection_for_each_tag(self, dataset_splits: Union[List[int], None]
                                       ) -> List[MoaSurveyLightCurveCollection]:
        """
        Creates a light curve collection for each tag in the survey and assigns the appropriate labels.

        :param dataset_splits: The dataset splits to include in each collection.
        :return: The list of collections.
        """
        collections = []
        for tag in self.moa_data_interface.survey_tag_to_path_list_dictionary.keys():
            if tag == 'i':
                continue  # There is a single `i` tag, which seems to be a typo.
            collection = MoaSurveyLightCurveCollection(survey_tags=[tag], dataset_splits=dataset_splits)
            if tag == 'cb':
                collection.label = [1, 1]
            elif tag in ['c', 'cf', 'cp', 'cw', 'cs']:
                collection.label = [1, 0]
            else:
                collection.label = [0, 0]
            collections.append(collection)
        return collections



"""
Code to download and prepare the data needed for the quick start tutorial.
"""
from pathlib import Path

from ramjet.data_interface.tess_data_interface import download_two_minute_cadence_light_curves
from ramjet.data_interface.tess_toi_data_interface import TessToiDataInterface
from ramjet.photometric_database.setup.tess_two_minute_cadence_transit_metadata import build_tables


def download_and_build_tables():
    """
    Downloads and prepares the data needed for the quick start tutorial.
    """
    download()
    build_tables()


def download():
    download_two_minute_cadence_light_curves(Path('data/tess_two_minute_cadence_light_curves'),
                                             limit=10000)
    tess_toi_data_interface = TessToiDataInterface()
    tess_toi_data_interface.download_exofop_toi_light_curves_to_directory(
        Path('data/tess_two_minute_cadence_light_curves'))


if __name__ == '__main__':
    download_and_build_tables()



"""
Code to represent a TESS light curve.
"""
from __future__ import annotations

import copy
from typing import Union, Optional, List

import lightkurve
import numpy as np

import pandas as pd
from astropy import units
from astropy.coordinates import SkyCoord, Angle
from astroquery.mast import Catalogs
from lightkurve import SearchResult
from lightkurve.targetpixelfile import TargetPixelFile
from retrying import retry

from ramjet.data_interface.tess_data_interface import is_common_mast_connection_error
from ramjet.photometric_database.light_curve import LightCurve


class TessLightCurve(LightCurve):
    """
    A class to represent a TESS light curve.
    """

    def __init__(self):
        super().__init__()
        self.tic_id: Union[int, None] = None
        self.sector: Union[int, None] = None
        self._tic_row: Union[None, pd.Series, MissingTicRow] = None

    @retry(retry_on_exception=is_common_mast_connection_error, stop_max_attempt_number=10)
    def get_tic_row(self):
        if self._tic_row is None:
            self._tic_row = Catalogs.query_object(f'TIC{self.tic_id}', catalog='TIC').to_pandas().iloc[0]
        return self._tic_row

    @property
    def sky_coord(self) -> SkyCoord:
        tic_row = self.get_tic_row()
        sky_coord = SkyCoord(ra=tic_row['ra'], dec=tic_row['dec'], unit=units.deg)
        return sky_coord

    @property
    def tess_magnitude(self) -> float:
        tic_row = self.get_tic_row()
        return float(tic_row['Tmag'])

    @retry(retry_on_exception=is_common_mast_connection_error, stop_max_attempt_number=10)
    def get_ffi_time_series_from_tess_cut(self) -> TargetPixelFile:
        search_result: SearchResult = lightkurve.search_tesscut(f'TIC{self.tic_id}', sector=self.sector)
        target_pixel_file = search_result.download(cutout_size=10)
        return target_pixel_file

    def estimate_photometric_centroid_of_variability_from_tess_ffi(self, minimum_period: Optional[float] = None,
                                                                   maximum_period: Optional[float] = None) -> SkyCoord:
        fold_period, fold_epoch, time_bin_size, minimum_bin_phase, maximum_bin_phase = \
            self.get_variability_phase_folding_parameters(minimum_period=minimum_period, maximum_period=maximum_period)
        variability_centroid_and_frames = \
            self.estimate_photometric_variability_centroid_and_frames_from_ffi_based_on_folding_parameters(
                fold_epoch, fold_period, maximum_bin_phase, minimum_bin_phase, time_bin_size)
        centroid_sky_coord = variability_centroid_and_frames[0]
        return centroid_sky_coord

    def estimate_photometric_variability_centroid_and_frames_from_ffi_based_on_folding_parameters(
            self, fold_epoch: float, fold_period: float, maximum_bin_phase: float, minimum_bin_phase: float,
            time_bin_size: float):
        target_pixel_file = self.get_ffi_time_series_from_tess_cut()
        if target_pixel_file is None:
            raise CentroidAlgorithmFailedError
        phases = ((target_pixel_file.time.value - fold_epoch) % fold_period)
        minimum_bin_indexes = np.where((phases > (minimum_bin_phase - time_bin_size)) &
                                       (phases < (minimum_bin_phase + time_bin_size)))
        maximum_bin_indexes = np.where((phases > (maximum_bin_phase - time_bin_size)) &
                                       (phases < (maximum_bin_phase + time_bin_size)))
        # Hack to get a single frame target pixel file with the right coordinates, etc. by copying the first frame.
        median_minimum_target_pixel_frame = copy.deepcopy(target_pixel_file[minimum_bin_indexes][0])
        median_minimum_target_pixel_frame.hdu[1].data["FLUX"] = np.nanmedian(
            target_pixel_file[minimum_bin_indexes].flux.value, axis=0, keepdims=True)
        # Hack to get a single frame target pixel file with the right coordinates, etc. by copying the first frame.
        median_maximum_target_pixel_frame = copy.deepcopy(target_pixel_file[maximum_bin_indexes][0])
        median_maximum_target_pixel_frame.hdu[1].data["FLUX"] = np.nanmedian(
            target_pixel_file[maximum_bin_indexes].flux.value, axis=0, keepdims=True)
        # Hack to get a single frame target pixel file with the right coordinates, etc. by copying the first frame.
        difference_target_pixel_frame = copy.deepcopy(target_pixel_file[0])
        difference_flux_frame = (median_maximum_target_pixel_frame.flux.value -
                                 median_minimum_target_pixel_frame.flux.value)
        try:
            difference_target_pixel_frame.hdu[1].data["FLUX"] = difference_flux_frame
        except ValueError as error:  # TODO: I don't know why this would happen here, but it is. Fix it.
            raise CentroidAlgorithmFailedError from error
        image_side_size = 10
        pixel_side_indexes = np.arange(image_side_size, dtype=np.float32)
        flux_difference = difference_flux_frame[0]
        positive_flux_difference = np.maximum(flux_difference, 0)
        try:
            x_flux_difference_centroid = np.average(pixel_side_indexes,
                                                    weights=np.mean(positive_flux_difference, axis=0))
            y_flux_difference_centroid = np.average(pixel_side_indexes,
                                                    weights=np.mean(positive_flux_difference, axis=1))
            centroid_sky_coord = target_pixel_file.wcs.pixel_to_world(x_flux_difference_centroid,
                                                                      y_flux_difference_centroid)
        except ZeroDivisionError as error:
            raise CentroidAlgorithmFailedError from error
        return (centroid_sky_coord, target_pixel_file, difference_target_pixel_frame, median_maximum_target_pixel_frame,
                median_minimum_target_pixel_frame)

    def estimate_angular_distance_to_variability_photometric_centroid_from_ffi(
            self, minimum_period: Optional[float] = None, maximum_period: Optional[float] = None) -> Angle:
        centroid_sky_coord = self.estimate_photometric_centroid_of_variability_from_tess_ffi(
            minimum_period=minimum_period,
            maximum_period=maximum_period)
        return self.sky_coord.separation(centroid_sky_coord)

    @classmethod
    def load_tic_rows_from_mast_for_list(cls, light_curves: List[TessLightCurve]) -> None:
        light_curve_tic_ids: List[int] = [light_curve.tic_id for light_curve in light_curves]
        tic_row_data_frame = Catalogs.query_criteria(ID=light_curve_tic_ids, catalog='TIC').to_pandas()
        for light_curve in light_curves:
            light_curve_tic_row_data_frame = tic_row_data_frame[tic_row_data_frame['ID'] == str(light_curve.tic_id)]
            if light_curve_tic_row_data_frame.shape[0] == 0:
                light_curve._tic_row = MissingTicRow
            else:
                light_curve._tic_row = light_curve_tic_row_data_frame.iloc[0]


class MissingTicRow:
    pass


class CentroidAlgorithmFailedError(Exception):
    pass



from ramjet.photometric_database.derived.tess_ffi_eclipsing_binary_light_curve_collection import \
    TessFfiAntiEclipsingBinaryForTransitLightCurveCollection
from ramjet.photometric_database.derived.tess_ffi_light_curve_collection import TessFfiLightCurveCollection
from ramjet.photometric_database.derived.tess_ffi_transit_light_curve_collections import \
    TessFfiConfirmedTransitLightCurveCollection, TessFfiNonTransitLightCurveCollection
from ramjet.photometric_database.standard_and_injected_light_curve_database import StandardAndInjectedLightCurveDatabase
from ramjet.photometric_database.light_curve_dataset_manipulations import OutOfBoundsInjectionHandlingMethod


class TessFfiDatabase(StandardAndInjectedLightCurveDatabase):
    """
    An abstract database with settings preset for the FFI data.
    """
    def __init__(self):
        super().__init__()
        self.batch_size = 1000
        self.time_steps_per_example = 1000
        self.shuffle_buffer_size = 100000
        self.out_of_bounds_injection_handling = OutOfBoundsInjectionHandlingMethod.RANDOM_INJECTION_LOCATION


magnitude_range = (0, 11)


class TessFfiStandardTransitDatabase(TessFfiDatabase):
    """
    A database using standard positive and negative transit light curves.
    """
    def __init__(self):
        super().__init__()
        self.training_standard_light_curve_collections = [
            TessFfiConfirmedTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range),
            TessFfiNonTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range)
        ]
        self.validation_standard_light_curve_collections = [
            TessFfiConfirmedTransitLightCurveCollection(dataset_splits=[8], magnitude_range=magnitude_range),
            TessFfiNonTransitLightCurveCollection(dataset_splits=[8], magnitude_range=magnitude_range)
        ]


class TessFfiStandardTransitAntiEclipsingBinaryDatabase(TessFfiDatabase):
    """
    A database using standard positive and negative transit light curves and a negative eclipsing binary collection.
    """
    def __init__(self):
        super().__init__()
        self.training_standard_light_curve_collections = [
            TessFfiConfirmedTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range),
            TessFfiNonTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range),
            TessFfiAntiEclipsingBinaryForTransitLightCurveCollection(dataset_splits=list(range(8)),
                                                                     magnitude_range=magnitude_range)
        ]
        self.validation_standard_light_curve_collections = [
            TessFfiConfirmedTransitLightCurveCollection(dataset_splits=[8], magnitude_range=magnitude_range),
            TessFfiNonTransitLightCurveCollection(dataset_splits=[8], magnitude_range=magnitude_range),
            TessFfiAntiEclipsingBinaryForTransitLightCurveCollection(dataset_splits=[8],
                                                                     magnitude_range=magnitude_range)
        ]
        self.inference_light_curve_collections = [
            TessFfiLightCurveCollection(dataset_splits=[9], magnitude_range=magnitude_range)]


class TessFfiInjectedTransitDatabase(TessFfiDatabase):
    """
    A database using positive and negative transit light curves injected into negative light curves.
    """
    def __init__(self):
        super().__init__()
        self.training_injectee_light_curve_collection = TessFfiNonTransitLightCurveCollection(
            dataset_splits=list(range(8)))
        self.training_injectable_light_curve_collections = [
            TessFfiConfirmedTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range),
            TessFfiNonTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range)
        ]
        self.validation_standard_light_curve_collections = [
            TessFfiConfirmedTransitLightCurveCollection(dataset_splits=[8], magnitude_range=magnitude_range),
            TessFfiNonTransitLightCurveCollection(dataset_splits=[8], magnitude_range=magnitude_range)
        ]


class TessFfiStandardAndInjectedTransitDatabase(TessFfiDatabase):
    """
    A database using positive and negative transit light curves injected into negative light curves.
    """
    def __init__(self):
        super().__init__()
        self.training_standard_light_curve_collections = [
            TessFfiConfirmedTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range),
            TessFfiNonTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range)
        ]
        self.training_injectee_light_curve_collection = TessFfiNonTransitLightCurveCollection(
            dataset_splits=list(range(8)), magnitude_range=magnitude_range)
        self.training_injectable_light_curve_collections = [
            TessFfiConfirmedTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range),
            TessFfiNonTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range)
        ]
        self.validation_standard_light_curve_collections = [
            TessFfiConfirmedTransitLightCurveCollection(dataset_splits=[8], magnitude_range=magnitude_range),
            TessFfiNonTransitLightCurveCollection(dataset_splits=[8], magnitude_range=magnitude_range)
        ]


class TessFfiStandardAndInjectedTransitAntiEclipsingBinaryDatabase(TessFfiDatabase):
    """
    A database using positive and negative transit light curves injected into negative light curves.
    """
    def __init__(self):
        super().__init__()
        self.training_standard_light_curve_collections = [
            TessFfiConfirmedTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range),
            TessFfiNonTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range),
            TessFfiAntiEclipsingBinaryForTransitLightCurveCollection(dataset_splits=list(range(8)),
                                                                     magnitude_range=magnitude_range)
        ]
        self.training_injectee_light_curve_collection = TessFfiNonTransitLightCurveCollection(
            dataset_splits=list(range(8)), magnitude_range=magnitude_range)
        self.training_injectable_light_curve_collections = [
            TessFfiConfirmedTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range),
            TessFfiNonTransitLightCurveCollection(dataset_splits=list(range(8)), magnitude_range=magnitude_range),
            TessFfiAntiEclipsingBinaryForTransitLightCurveCollection(dataset_splits=list(range(8)),
                                                                     magnitude_range=magnitude_range)
        ]
        self.validation_standard_light_curve_collections = [
            TessFfiConfirmedTransitLightCurveCollection(dataset_splits=[8], magnitude_range=magnitude_range),
            TessFfiNonTransitLightCurveCollection(dataset_splits=[8], magnitude_range=magnitude_range),
            TessFfiAntiEclipsingBinaryForTransitLightCurveCollection(dataset_splits=list(range(8)),
                                                                     magnitude_range=magnitude_range)
        ]



from copy import deepcopy
from dataclasses import dataclass
from typing import Self

import numpy as np
import numpy.typing as npt
from public import public


@public
@dataclass
class LightCurve:
    """
    A class to represent a light curve. A light curve is a collection of data which may includes times, fluxes,
    and related values.

    :ivar times: The times of the light curve.
    :ivar fluxes: The fluxes of the light curve.
    """
    times: npt.NDArray[np.float32]
    fluxes: npt.NDArray[np.float32]

    @classmethod
    def new(cls, times: npt.NDArray[np.float32], fluxes: npt.NDArray[np.float32]) -> Self:
        """
        Creates a new light curve.

        :param times: The times of the light curve.
        :param fluxes: The fluxes of the light curve.
        :return: The light curve.
        """
        return cls(times=times, fluxes=fluxes)


def remove_nan_flux_data_points_from_light_curve(light_curve: LightCurve) -> LightCurve:
    light_curve = deepcopy(light_curve)
    nan_flux_indexes = np.isnan(light_curve.fluxes)
    light_curve.fluxes = light_curve.fluxes[~nan_flux_indexes]
    light_curve.times = light_curve.times[~nan_flux_indexes]
    return light_curve



"""
Code for representing a collection of light curves.
"""
import numpy as np
from pathlib import Path
from typing import Iterable, Union, List, Optional


class LightCurveCollectionMethodNotImplementedError(RuntimeError):
    """
    An error to raise if a collection method that is not implemented is attempted to be used.
    Note, the standard NotImplementedError is not supposed to be used for cases when non-implemented functions are
    meant to be allowed, which is why a custom class is needed.
    """
    pass


class LightCurveCollection:
    """
    A class representing a collection of light curves. Used to define how to find, load, and label a set of light 
    curves.

    :ivar label: The default label to be used if the `load_label_from_path` method is not overridden.
    :ivar paths: The default list of paths to be used if the `get_paths` method is not overridden.
    """
    def __init__(self):
        self.label: Union[float, List[float], np.ndarray, None] = None
        self.paths: Union[List[Path], None] = None

    def get_paths(self) -> Iterable[Path]:
        """
        Gets the paths for the light curves in the collection.

        :return: An iterable of the light curve paths.
        """
        return self.paths

    def load_times_and_fluxes_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and fluxes from a given light curve path.

        :param path: The path to the light curve file.
        :return: The times and the fluxes of the light curve.
        """
        raise LightCurveCollectionMethodNotImplementedError

    def load_times_and_magnifications_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and magnifications from a given path as an injectable signal.

        :param path: The path to the light curve/signal file.
        :return: The times and the magnifications of the light curve/signal.
        """
        raise LightCurveCollectionMethodNotImplementedError

    @staticmethod
    def generate_synthetic_signal_from_real_data(fluxes: np.ndarray, times: np.ndarray) -> (np.ndarray, np.ndarray):
        """
        Takes real light curve data and converts it to a form that can be used for synthetic light curve injection.

        :param fluxes: The real light curve fluxes.
        :param times: The real light curve times.
        :return: Fake synthetic magnifications and times.
        """
        fluxes -= np.minimum(np.min(fluxes), 0)  # Fix negative flux cases if they exist.
        flux_median = np.median(fluxes)
        normalized_fluxes = fluxes / flux_median
        relative_times = times - np.min(times)
        return normalized_fluxes, relative_times

    def load_label_from_path(self, path: Path) -> Union[float, np.ndarray]:
        """
        Loads the label of an example from a corresponding path.

        :param path: The path to load the label for.
        :return: The label.
        """
        return self.label

    @staticmethod
    def shuffle_and_split_paths(paths: List[Path], dataset_splits: List[int], number_of_splits: int = 10) -> List[Path]:
        """
        Repeatably shuffles a list of paths and then gets the requested dataset splits from that list of paths.
        Designed to allow splitting a list of paths into training, validation, and testing datasets easily.

        :param paths: The original list of paths.
        :param dataset_splits: The indexes of the dataset splits to return.
        :param number_of_splits: The number of dataset splits.
        :return: The paths of the dataset splits.
        """
        path_array = np.array(paths)
        np.random.seed(0)
        np.random.shuffle(path_array)
        dataset_split_arrays = np.array_split(path_array, number_of_splits)
        dataset_split_arrays_to_keep = [dataset_split_array
                                        for dataset_split_index, dataset_split_array in enumerate(dataset_split_arrays)
                                        if dataset_split_index in dataset_splits]
        paths_array = np.concatenate(dataset_split_arrays_to_keep)
        dataset_split_paths = list(map(Path, paths_array))
        return dataset_split_paths

    def load_times_fluxes_and_flux_errors_from_path(self, path: Path
                                                    ) -> (np.ndarray, np.ndarray, Union[np.ndarray, None]):
        """
        Loads the times, fluxes, and flux errors of a light curve from a path to the data.
        Unless overridden, defaults to using the method to load only the times and fluxes, and returns None for errors.

        :param path: The path of the file containing the light curve data.
        :return: The times, fluxes, and flux errors.
        """
        times, fluxes = self.load_times_and_fluxes_from_path(path)
        flux_errors = None
        return times, fluxes, flux_errors

    def load_times_magnifications_and_magnification_errors_from_path(
            self, path: Path) -> (np.ndarray, np.ndarray, Union[np.ndarray, None]):
        """
        Loads the times, magnifications, and magnification_errors of a light curve from a path to the data.
        Unless overridden, defaults to using the method to load only the times and magnifications,
        and returns None for magnification errors.

        :param path: The path of the file containing the light curve data.
        :return: The times, magnifications, and magnification errors.
        """
        times, fluxes = self.load_times_and_magnifications_from_path(path)
        flux_errors = None
        return times, fluxes, flux_errors

    def load_auxiliary_information_for_path(self, path: Path) -> np.ndarray:
        """
        Loads auxiliary information for the given path.

        :param path: The path to the light curve file.
        :return: The auxiliary information.
        """
        return np.array([], dtype=np.float32)
